#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
Check validity of json-schemas themselves.
'''
import unittest

import jsonschema

from .. import datamodel


class Test(unittest.TestCase):

    @unittest.expectedFailure #Due to extra types: DataFrame, Series
    def testSchema(self):
        validator = datamodel.model_validator()
        validator.check_schema(datamodel.model_schema())

    def testShema_emptyInstanceFail(self):
        validator = datamodel.model_validator()
        instance = {}

        self.assertRaises(jsonschema.ValidationError, validator.validate, instance)

    def testModelBase_fail(self):
        mdl = datamodel.base_model()

        self.assertRaises(jsonschema.ValidationError, datamodel.model_validator().validate, mdl)

        datamodel.set_jsonpointer(mdl, '/engine/fuel', 'BAD_FUEL')
        self.assertRaisesRegex(jsonschema.ValidationError, "Failed validating 'oneOf' in schema.*properties.*engine", datamodel.model_validator().validate, mdl)


    def testModel_FAIL_extraFuel(self):
        mdl = datamodel.base_model()
        datamodel.set_jsonpointer(mdl, '/engine/fuel', 'diesel')
        mdl['params']['fuel']['EXTRA_FUEL'] = 'somethign'

        self.assertRaisesRegex(jsonschema.ValidationError, "Additional properties .*EXTRA_FUEL", datamodel.model_validator().validate, mdl)

    def testModel_FAIL_missLhv(self):
        mdl = datamodel.base_model()
        datamodel.set_jsonpointer(mdl, '/engine/fuel', 'diesel')
        mdl['params']['fuel']['petrol'] = {}

        self.assertRaisesRegex(jsonschema.ValidationError, "'lhv' is a required", datamodel.model_validator().validate, mdl)


    def testModel_GOOD(self):
        mdl = datamodel.base_model()
        datamodel.set_jsonpointer(mdl, '/engine/fuel', 'diesel')

        validator = datamodel.model_validator()
        validator.validate(mdl)

    def testModel_units_GOOD(self):
        from pint import UnitRegistry
        ureg = UnitRegistry()

        cases = [
            [['/engine/bore', None], 0.14 * ureg.meter],
            [['/engine/bore', 0.14], 0.14 * ureg.meter],
            [['/engine/bore', '0.14 (m)'], 0.14 * ureg.meter],
            [['/engine/bore', '+0.14 (m)'], 0.14 * ureg.meter],
            [['/engine/bore', '14 (mm)'], 0.14 * ureg.meter],
            [['/engine/bore', '14'], 0.14 * ureg.meter],
            [['/engine/bore', ' +14 '], 0.14 * ureg.meter],
        ]

        validator = datamodel.model_validator()

        for (args, res) in cases:
            mdl = datamodel.base_model()
            datamodel.set_jsonpointer(mdl, '/engine/fuel', 'diesel')

            datamodel.set_jsonpointer(mdl, *args)
            validator.validate(mdl)




if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#! python3
# -*- coding: utf-8 -*-
#
# Sphinx documentation build configuration file.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os, io


projname = 'fuefit'
mydir = os.path.dirname(__file__)

def read_project_version():
    fglobals = {}
    with io.open(os.path.join(mydir, '..', projname, '_version.py')) as fd:
        exec(fd.read(), fglobals)  # To read __version__
    return fglobals['__version__']

proj_ver = read_project_version()

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../'))
#sys.path.insert(0, os.path.abspath('../devtools')) # Does not work for scripts :-(


## Mock C-libraries (numpy/pandas, etc) so that `autodoc` sphinx-extension
#    can import sources.
#     From http://read-the-docs.readthedocs.org/en/latest/faq.html#i-get-import-errors-on-libraries-that-depend-on-c-modules
#
if on_rtd:
    from unittest.mock import MagicMock

    class Mock(MagicMock):
        @classmethod
        def __getattr__(cls, name):
            if name in ('__file__', '__path__'):
                return '/dev/null'
            elif name[0] == name[0].upper():
                mockType = type(name, (), {})
                mockType.__module__ = __name__
                return mockType
            else:
                return Mock()

    MOCK_MODULES = ['matplotlib', 'matplotlib.sphinxext', 'matplotlib.sphinxext.plot_directive', 'xlwings']
    sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

## Trick from https://github.com/rtfd/readthedocs.org/issues/283
# On read the docs we need to use a different CDN URL for MathJax which loads
# over HTTPS.
if on_rtd:
    mathjax_path = 'https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML'

## Make autodoc always includes constructors.
#    From http://stackoverflow.com/a/5599712/548792
#    and http://stackoverflow.com/questions/3757500/how-do-i-connect-sphinxs-autodoc-skip-member-to-my-function
#
autodoc_default_flags = ['members', 'private-members',
                    'special-members',
                    #'undoc-members',
                    'show-inheritance']
# autoclass_content = 'both' ## Join class+ __init__() docstrings
def autodoc_skip_member(app, what, name, obj, skip, options):
    exclusions = ('__weakref__',  # special-members
                  '__doc__', '__module__', '__dict__',  # undoc-members
                  #'__init__', __new__
                  )
    exclude = name in exclusions
    return skip or exclude
# #     if what == 'method' and \
# #             (name == "__init__" or re.match(r'^_[A-Za-z]', name)):
#     if (name == "__init__" or re.match(r'^_[A-Za-z]', name)):
#         return False
#     return skip

def setup(app):
    app.connect("autodoc-skip-member", autodoc_skip_member)


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'matplotlib.sphinxext.plot_directive',
]

## Prevent Sphinx from doctesting python-modules doctests,
# they will be checked separately with `nose` plugin.
doctest_test_doctest_blocks = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'fuefit'
copyright = '2013-2014, European Commission (JRC), EUPL 1.1+'  # @ReservedAssignment

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = proj_ver
# The full version, including alpha/beta/rc tags.
release = proj_ver

extlinks = {
    'issue': ('https://github.com/ankostis/fuefit/issues/%s','issue')
}
todo_include_todos = True

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', 'docs']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'obj'

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
if not on_rtd:
    import sphinx_rtd_theme  # @UnresolvedImport
    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

    html_theme_options = {
        'sticky_navigation':True,
    }

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {
# }

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'fuefitdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',
'papersize': 'a4paper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
'preamble': u'''
    \\usepackage{amsmath}
''',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  ('index', 'fuefit.tex', 'fuefit Documentation',
   'Kostis Anagnostopoulos', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'fuefit', 'fuefit Documentation',
     ['Kostis Anagnostopoulos'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'fuefit', 'fuefit Documentation',
   'Kostis Anagnostopoulos', 'fuefit', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('http://docs.python.org/3.4', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference/', None),
    'pandas': ('http://pandas.pydata.org/pandas-docs/dev', None),
    'lmfit': ('http://lmfit.github.io/lmfit-py/', None),
    'jsonschema': ('http://python-jsonschema.readthedocs.org/en/latest/', None),
}




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
import io
from textwrap import dedent
from collections import OrderedDict
'''
Function for building and jsonschema-validating input and output model.
'''

from collections.abc import Mapping, Sequence 
import json

import numpy as np
from pandas.core.generic import NDFrame

import jsonschema as jsons
import operator as ops
import pandas as pd


def model_schema(additional_properties = False):
    """The json-schema for input/output of the fuefit experiment.

    :return :dict:
    """

    from textwrap import dedent

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": model_schema.__doc__,
        "type": "object", "additionalProperties": additional_properties,
        "required": ["engine"],
        "properties": {
            "engine": {
                "oneOf": [{
                        "title": "engine series",
                        "type": "Series",
                    },
                    {
                        "title": "engine model",
                        "type": "object", "additionalProperties": additional_properties,
                        "required": ["fuel"],
                        "description": "The engine attributes and  data-points vectors required for generating a fitted fuel-map.",
                        "properties": {
                            "fuel": {
                                "title": "fuel (petrol | diesel)",
                                "enum": [ "petrol", "diesel" ],
                                "description": "the engine's fuel-type used for selecting specific-temperature and/or load-curve.",
                            },
                            "p_max": {
                                "title": "maximum rated power",
                                "$ref": "#/definitions/positiveQuantityOrNumOrNull",
                                "description": dedent("""
                                    The maximum rated engine power as declared by the manufacturer.
                                    Required if `p_norm` or `fc_norm` exists in input-file's or example-map's columns.""")
                            },
                            "n_rated": {
                                "title": "rated engine revolutions (rad/min)",
                                "$ref": "#/definitions/positiveQuantityOrNumOrNull",
                                "description": dedent("""
                                    The engine's revolutions where maximum-power is attained.
                                    Required if `n_norm` exists in input-file's or example-map columns."""),
                            },
                            "n_idle": {
                                "title": "idling revolutions (rad/min)",
                                "$ref": "#/definitions/positiveQuantityOrNumOrNull",
                                "description": dedent("""
                                    The engine's revolutions when idle.
                                    Required if `n_norm` exists in input-file's or example-map columns."""),
                            },
                            'stroke': {
                                "title": "piston stroke (mm)",
                                "$ref": "#/definitions/positiveQuantityOrNumOrNull",
                                'description': dedent("""
                                    The engine's stroke traveling distance.
                                    Required if CM is not among the inputs or requested to generate example-map with RPM column.""")
                            },
                            'capacity': {
                                "title": "engine capacity (cm^3)",
                                "$ref": "#/definitions/positiveQuantityOrNumOrNull",
                                'description': dedent("""
                                    The total displacement of all engine's pistons.
                                    This value is ignored' if `stroke`, `bore` and `cylinders` are all present.
                                    Required if `pmf` is not among the inputs or requested to generate example-map with `fc` column.""")
                            },
                            'bore': {
                                "title": "piston bore (mm)",
                                "$ref": "#/definitions/positiveQuantityOrNumOrNull",
                                'description': dedent("""
                                    The piston diameter.
                                    The `capacity` is calculated from `stroke`, `bore` and `cylinders` when are all present.""")
                            },
                            'cylinders': {
                                "title": "number of cylinders (mm)",
                                "$ref": "#/definitions/positiveIntegerOrNull",
                                'description': dedent("""
                                    The number of cyclinders in the engine.
                                    The `capacity` is calculated from `stroke`, `bore` and `cylinders` when are all present.""")
                            },
                            'fuel_lhv': {
                                "title": "Fuel's Specific Heat-Value (kjoule/kgr)",
                                "$ref": "#/definitions/positiveNumber",
                                'description': dedent("""
                                    If set, overrides any value that would be selected from `/params/fuel/XXX` based on `/engine/fuel`. """)
                            },
                            'fc_map_coeffs': {
                                "title": "Fitted coefficients",
                                "type": "Series",
                                'description': dedent("""
                                    The result of the fitting: a, b, c, a2, b2, loss0, loss2
                                    """)
                            },
                        }
                    }  #engine-props
                ]
            }, #engine
            'measured_eng_points':{
                "type": "DataFrame"
            }, #measured_eng_points
            "params": {
                "title": "Experiment parameters and constants",
                "type": "object", "additionalProperties": additional_properties,
                "required": ['fuel', 'fitting'],
                "properties": {
                        'fuel': {
                            "title": "fuel-types",
                            "type": "object", "additionalProperties": additional_properties,
                            "required": ['petrol', 'diesel'],
                            "properties": {
                                'diesel': {
                                    "title": "typical diesel params",
                                    "$ref": "#/definitions/fuel_spec",
                                },
                                'petrol': {
                                    "title": "petrol params",
                                    "$ref": "#/definitions/fuel_spec",
                                },
                            } #fuel-props
                        }, #fuel
                        'plot_maps': {
                            "title": "Plot engine-maps?",
                            "type": ["boolean", "number"],
                            "default": False,
                        },
                        'is_robust': {
                            "title": "Robust fitting?",
                            "description": dedent("""
                                When `robust`, outliers are excluded from the fitted-data,
                                by using a non-linear iteratively-reweighted least-squares (IRLS) fitting-method.
                                
                                See :func:`fuefit.robustfit.curve_fit()` for more.
                            """),
                            "type": ["boolean", "null"],
                            "default": False,
                        },
                        'fitting': {
                            "type": "object",
                            "properties": {
                                'coeffs': {
                                    "title": "Fitting Coefficient params",
                                },
                                'fit_options': {
                                    "title": "Least-square fitting options",
                                },
                            },
                        }
                }
            },
        },
        "definitions": {
            "positiveInteger": {
                "type": "integer",
                "minimum": 0,
                "exclusiveMinimum": True,
            },
            "positiveIntegerOrNull": {
                "type": ["integer", 'null'],
                "minimum": 0,
                "exclusiveMinimum": True,
            },
            "positiveNumber": {
                "type": "number",
                "minimum": 0,
                "exclusiveMinimum": True,
            },
            "positiveNumberOrNull": {
                "type": ["number", 'null'],
                "minimum": 0,
                "exclusiveMinimum": True,
            },
            "positiveIntegers": {
                "type": "array",
               "items": { "$ref": "#/definitions/positiveInteger" },
            },
            "positiveNumbers": {
                "type": "array",
               "items": { "$ref": "#/definitions/positiveNumber" },
            },
            "numbers": {
                "type": "array",
               "items": { "type": "number" },
            },

            "quantity": {
                "type": "string",
                "pattern": "^\s*[-+]?\s*\d+(\.\d*)?\s*(\([^)]+\))?\s*$",
            },
            "positiveQuantity": {
                "type": "string",
                "pattern": "^\s*\+?\s*\d+(\.\d*)?\s*(\([^)]+\))?\s*$",
            },
            "negativeQuantity": {
                "type": "string",
                "pattern": "^\s*-\s*\d+(\.\d*)?\s*(\([^)]+\))?\s*$",
            },
            "positiveQuantityOrNumOrNull": {
                "oneOf": [
                    { "$ref": "#/definitions/positiveQuantity" },
                    { "$ref": "#/definitions/positiveNumber" },
                    { "type": "null" },
                ],
            },

            "fuel_spec": {
                "type": "object",
                "required": ['lhv'],
                "properties": {
                    'lhv': {'title': "Fuel's Specific Heat-Value (kjoule/kgr)", "$ref": "#/definitions/positiveInteger"}
                }
            },
            "mergeableArray": {
                "type": "object", "": False,
                "required": ["$merge", "$list"],
                "properties": {
                    "$merge": {
                        "enum": ["merge", "replace", "append_head", "append_tail", "overwrite_head", "overwrite_tail"],
                        "description": dedent("""
                            merge       := appends any non-existent elements
                            replace     := (default) all items replaced
                        """),
                    },
                    "$list": {
                        "type": "array",
                    }
                },
            },
            "mergeableObject": {
                "type": "object",
                "properties": {
                    "$merge": {
                        "type": "boolean",
                        "description": dedent("""
                            true    := (default) merge properties
                            false   := replace properties
                        """),
                    },
                },
            },
        }
    }

    return schema



def model_validator(additional_properties=False):
    from jsonschema import Draft4Validator
    schema = model_schema(additional_properties)
    validator = Draft4Validator(schema)
    validator._types.update({"object": (dict, pd.Series, pd.DataFrame), "DataFrame" : pd.DataFrame, 'Series':pd.Series})

    return validator

def validate_model(mdl, additional_properties=False):
    validator = model_validator(additional_properties=False)
    try:
        validator.validate(mdl)
    except jsons.ValidationError as ex:
        ## Attempt to workround BUG: https://github.com/Julian/jsonschema/issues/164
        #
        if isinstance(ex.instance, NDFrame):
            ex.instance = str(ex.instance)
        raise



def base_model():
    '''The base model for running a WLTC experiment.

    It contains some default values for the experiment (ie the default 'full-load-curve' for the vehicles).
    But note that it this model is not valid - you need to override its attributes.

    :return :json_tree: with the default values for the experiment.
    '''

    instance = {
        'engine': {
            "fuel":     None,
            "p_max":  None,
            "n_idle":   None,
            "n_rated":  None,
            "stroke":    None,
            "capacity":    None,
            "bore":    None,
            "cylinders":    None,
        },
        'params': {
            'fuel': {
                'diesel':   {'lhv':42700},
                'petrol':   {'lhv':43000},
            },
            'fitting': {
                'is_robust':        False,
                'fit_options': {},
                'coeffs': OrderedDict([
                    ('a',     dict(value=0.45)),
                    ('b',     dict(value=0.0154)),
                    ('c',     dict(value=-0.00093)),
                    ('a2',    dict(value=-0.0027)),
                    ('b2',    dict(value=0)),
                    ('loss0', dict(value=-2.17)),
                    ('loss2', dict(value=-0.0037)),
                ]), 
            },
            'plot_maps':        False,
        }
    }

    return instance


def make_json_defaulter(pd_method):
    def defaulter(o):
        if (isinstance(o, NDFrame)):
            if pd_method is None:
                s = json.loads(pd.DataFrame.to_json(o))
            else:
                method = ops.methodcaller(pd_method)
                s = '%s:%s'%(type(o).__name__, method(o))
        else:
            s =repr(o)
        return s

    return defaulter

def json_dumps(obj, pd_method=None):
    return json.dumps(obj, indent=2, default=make_json_defaulter(pd_method))
def json_dump(obj, fp, pd_method=None):
    json.dump(obj, fp, indent=2, default=make_json_defaulter(pd_method))


try:
    from enum import Enum       # @UnresolvedImport @UnusedImport
except:
    from enum34 import Enum     # @UnresolvedImport @UnusedImport @Reimport

class MergeMode(Enum):
    REPLACE       = 1
    APPEND_HEAD   = 2
    APPEND_TAIL   = 3
    OVERLAP_HEAD  = 4
    OVERLAP_TAIL  = 5


def islist(obj):
    return isinstance(obj, Sequence) and not isinstance(obj, str)

def merge(a, b, path=[], list_merge_mode = MergeMode.REPLACE, raise_struct_mismatches = False):
    ''''Merges b into a.

    List merge modes: REPLACE, APPEND_HEAD, APPEND_TAIL, OVERLAP_HEAD, OVERLAP_TAIL
    '''

    def issue_struct_mismatch(mismatch_type, key, a_value, b_value):
        raise ValueError("%s-values conflict at '%s'! a(%s) != b(%s)" %
                                (mismatch_type, '/'.join(path + [str(key)]), type(av), type(bv)))

    for key in b:
        bv = b[key]
        if key in a:
            av = a[key]
            if av == bv:
                continue # same leaf value

            if raise_struct_mismatches:
                if isinstance(av, Mapping) != isinstance(bv, Mapping):
                    issue_struct_mismatch('Dict', key, av, bv)
                elif islist(av) != islist(bv):
                    issue_struct_mismatch('List', key, av, bv)

            if isinstance(av, Mapping):
                merge(av, bv, path + [str(key)])
                continue

            if islist(av) and list_merge_mode > MergeMode.REPLACE:
                # TODO: merge/append lists
                #continue
                pass

        a[key] = bv
    return a

# works
# print(merge({1:{"a":"A"},2:{"b":"B"}}, {2:{"c":"C"},3:{"d":"D"}}))
# # has conflict
# merge({1:{"a":"A"},2:{"b":"B"}}, {1:{"a":"A"},2:{"b":"C"}})



class JsonPointerException(Exception):
    pass


def jsonpointer_parts(jsonpointer):
    """
    Iterates over the ``jsonpointer`` parts.

    :param str jsonpointer: a jsonpointer to resolve within document
    :return: a generator over the parts of the json-pointer

    :author: Julian Berman, ankostis
    """

    if jsonpointer:
        parts = jsonpointer.split(u"/")
        if parts.pop(0) != '':
            raise JsonPointerException('Location must starts with /')
    
        for part in parts:
            part = part.replace(u"~1", u"/").replace(u"~0", u"~")
    
            yield part

_scream = object()
def resolve_jsonpointer(doc, jsonpointer, default=_scream):
    """
    Resolve a ``jsonpointer`` within the referenced ``doc``.
    
    :param doc: the referrant document
    :param str jsonpointer: a jsonpointer to resolve within document
    :return: the resolved doc-item or raises :class:`JsonPointerException` 

    :author: Julian Berman, ankostis
    """
    for part in jsonpointer_parts(jsonpointer):
        if isinstance(doc, Sequence):
            # Array indexes should be turned into integers
            try:
                part = int(part)
            except ValueError:
                pass
        try:
            doc = doc[part]
        except (TypeError, LookupError):
            if default is _scream:
                raise JsonPointerException(
                    "Unresolvable JSON pointer(%r)@(%s)" % (jsonpointer, part)
                )
            else:
                return default
        
    return doc

        
def set_jsonpointer(doc, jsonpointer, value, object_factory=dict):
    """
    Resolve a ``jsonpointer`` within the referenced ``doc``.
    
    :param doc: the referrant document
    :param str jsonpointer: a jsonpointer to the node to modify 
    :raises: JsonPointerException (if jsonpointer empty, missing, invalid-contet)
    """
    
    
    parts = list(jsonpointer_parts(jsonpointer))
        
    ## Will scream if used on 1st iteration.
    #
    pdoc = None
    ppart = None
    for i, part in enumerate(parts):
        if isinstance(doc, Sequence) and not isinstance(doc, str):
            ## Array indexes should be turned into integers
            #
            doclen = len(doc)
            if part == '-':
                part = doclen
            else:
                try:
                    part = int(part)
                except ValueError:
                    raise JsonPointerException("Expected numeric index(%s) for sequence at (%r)[%i]" % (part, jsonpointer, i))
                else:
                    if part > doclen:
                        raise JsonPointerException("Index(%s) out of bounds(%i) of (%r)[%i]" % (part, doclen, jsonpointer, i))
        try:
            ndoc = doc[part]
        except (LookupError):
            break  ## Branch-extension needed.
        except (TypeError): # Maybe indexing a string...
            ndoc = object_factory()
            pdoc[ppart] = ndoc
            doc = ndoc
            break  ## Branch-extension needed.
    
        doc, pdoc, ppart = ndoc, doc, part 
    else:
        doc = pdoc # If loop exhausted, cancel last assignment.

    ## Build branch with value-leaf.
    #
    nbranch = value
    for part2 in reversed(parts[i+1:]):
        ndoc = object_factory()
        ndoc[part2] = nbranch
        nbranch = ndoc
        
    ## Attach new-branch. 
    try:
        doc[part] = nbranch
    except IndexError: # Inserting last sequence-element raises IndexError("list assignment index out of range")
        doc.append(nbranch)
    
#    except (IndexError, TypeError) as ex:
#        #if isinstance(ex, IndexError) or 'list indices must be integers' in str(ex):
#        raise JsonPointerException("Incompatible content of JSON pointer(%r)@(%s)" % (jsonpointer, part))
#        else:
#            doc = {}
#            parent_doc[parent_part] = doc 
#            doc[part] = value 



def ensure_modelpath_Series(mdl, json_path):
    part = resolve_jsonpointer(mdl, json_path, None)
    if not isinstance(part, pd.Series):
        part = pd.Series(part)
        set_jsonpointer(mdl, json_path, part)

def ensure_modelpath_DataFrame(mdl, json_path):
    part = resolve_jsonpointer(mdl, json_path, None)
    if not isinstance(part, pd.Series):
        part = pd.DataFrame(part) 
        set_jsonpointer(mdl, json_path, part)



if __name__ == "__main__":
    print("Model: %s" % json.dumps(model_schema(), indent=2))




#!python
#-*- coding: utf-8 -*-
"""
Sample `xlwings` script
#######################
 
The functions included provide for running a batch of experiments in an excel-table 
with json-pointer paths as headers (see accompanying :file:`.xlsm`).

You can debug it by running it directly as a python script, as suggested by : 
  http://docs.xlwings.org/debugging.html

<< EDIT THIS SCRIPT TO PUT YOUR EXCEL/XLWINGS PYTHON-CODE, BELOW >>
"""



from __future__ import division, print_function, unicode_literals

import logging
import operator
import os
import re

from fuefit import datamodel
from fuefit import processor
import pandas as pd
import xlwings as xw


DEFAULT_LOG_LEVEL   = logging.INFO
def _init_logging(loglevel):
    logging.basicConfig(level=loglevel)
    logging.getLogger().setLevel(level=loglevel)

    log = logging.getLogger(__name__)
    log.trace = lambda *args, **kws: log.log(0, *args, **kws)
    
    return log
log = _init_logging(DEFAULT_LOG_LEVEL)
 
## TODO: Convert Excel-ref RC-notation to A1
_excel_ref_specifier_regex = re.compile(r'''^\s*
            @
            (?:(?P<sheet>.+)!)?             # Sheet-name optional-group
            (?P<ref>                        # start Cell-ref 
                (?:[A-Z]+\d+ | R\d+C\d+ | \(\d+,\d+\))        # FROM-ref, RC/A1 notation, or tuple
                (?:                             # start TO-Cell-ref optional-group
                    :
                    (?:[A-Z]+\d+ | R\d+C\d+ | \(\d+,\d+\))        # RC/A1 notation, or tuple
                )?                              # end TO-Cell-ref optional-group
            )                               # end Cell-ref
            (?:                             # start Shape-specifier optional-group
                \.
                (?P<shape>table|vertical|horizontal)     # See respective xw.Range attributes
            )?                              # end Shape-specifier
            (?:\((?P<range_kws>                   # start RANGE-kws expression
                [^)]*
            )\))?                            # end RANGE-kws
            (?:{(?P<pandas_kws>                   # start PANDAS-kws expression
                [^)]*
            )})?                            # end PANDAS-kws
            \s*$''', re.X+re.IGNORECASE)
_undefined = object()
def _parse_kws(kws_str):
    if kws_str:
        local_vars = {}
        exec('kws = dict(%s)' % kws_str, None, local_vars)
        return local_vars['kws']
    return {}

def resolve_excel_ref(ref_str, default=_undefined):
    """
    if `ref_str` is an *excel-ref*, it returns the referred-contents as DataFrame or a scalar, `None` otherwise.
    
    The *excel-ref* syntax is case-insensitive apart from the key-value pairs, 
    which it is given in BNF-notation:
    
    .. productionlist::
            excel_ref   : "@" 
                        : [sheet "!"] 
                        : cells 
                        : ["." shape] 
                        : ["(" range_kws ")"] 
                        : ["{" df_kws "}"]
            sheet       : sheet_name | sheet_index
            sheet_name  : <any character>
            sheet_index : `integer`
            cells       : cell_ref [":" cell_ref]
            cell_ref    : A1_ref | RC_ref | tuple_ref
            A1_ref      : <ie: "A1" or "BY34">
            RC_ref      : <ie: "R1C1" or "R24C34">
            tuple_ref   : <ie: "(1,1)" or "(24,1)", the 1st is the row>
            shape       : "." ("table" | "vertical" | "horizontal")
            range_kws   : kv_pairs                    # keywords for xlwings.Range(**kws)
            df_kws      : kv_pairs                    # keywords for pandas.DataFrafe(**kws)
            kv_pairs    : <python code for **keywords ie: "a=3.14, f = 'gg'">
            
    Excel-ref examples::
    
            @a1
            @E5.column
            @some sheet_name!R1C5.TABLE
            @1!a1:c5.table(header=False)
            @3!a1:C5.horizontal(strict=True; atleast_2d=True)
            @sheet-1!A1.table(asarray=True){columns=['a','b']}
            @any%sheet^&name!A1:A6.vertical(header=True)        ## NOTE: Setting Range's `header` kw and 
                                                                #      DataFrame will parse 1st row as header
    
    Note that the "RC-notation" is not converted, so Excel may not support it (unless overridden in its options).
    """
    
    matcher = _excel_ref_specifier_regex.match(ref_str)
    if matcher:
        ref = matcher.groupdict()
        log.info("Parsed string(%s) as Excel-ref: %s", ref_str, ref)
        
        sheet = ref.get('sheet') or ''
        try:
            sheet = int(sheet)
        except ValueError:
            pass
        cells = ref['ref']
        range_kws = _parse_kws(ref.get('range_kws'))
        ref_range = xw.Range(sheet, cells, **range_kws)
        range_shape = ref.get('shape')
        if range_shape:
            ref_range = operator.attrgetter(range_shape.lower())(ref_range)

        v = ref_range.value
        
        if ref_range.row1 != ref_range.row2 or ref_range.col1 != ref_range.col2:
            ## Parse as DataFrame when more than one cell.
            #
            pandas_kws = _parse_kws(ref.get('pandas_kws'))
            if 'header' in range_kws and not 'columns' in pandas_kws :
                ##Do not ignore explicit-columns.
                v = pd.DataFrame(v[1:], columns=v[0], **pandas_kws)
            else:
                v = pd.DataFrame(ref_range.value, **pandas_kws)
        else:
            v = ref_range.value
            
        log.debug("Excel-ref(%s) value: %s", ref, v)

        return v
    else:
        if default is _undefined:
            raise ValueError("Invalid excel-ref(%s)!" % ref_str)
        else:
            return default 
    
    
def read_input_as_df(table_ref_str):
    """
    Expects excel-table with jsonpointer paths as Header and 1st column named `id` as index, 
    like this::
    
        id     vehicle/test_mass  vehicle/p_rated           vehicle/gear_ratios 
        veh_1               1500              100  [120.75, 75, 50, 43, 37, 32] 
        veh_2               1600               80  [120.75, 75, 50, 43, 37, 32] 
        veh_3               1200               60  [120.75, 75, 50, 43, 37, 32] 
        
    """
    vehs = xw.Range('Input', table_ref_str, strict=True).table.value
    vehs = pd.DataFrame(vehs[1:], columns=vehs[0]).set_index('id')
    
    return vehs


def build_models(vehs_df):
    """
    Builds all input-dataframes as Experiment classes and returns them in a list of (veh_id, exp) pairs.
    
    :param vehs_df:     A dataframe indexed by veh_id, and with columns *json-pointer* paths into the model
    :return: a list of (veh_id, model) tuples
    """
    experiment_pairs = []
    for veh_id, row in vehs_df.iterrows():
        try:
            mdl_in = datamodel.base_model()
            for colname, colval in row.items():
                log.debug('veh_id(%s): Column(%s): %s', veh_id, colname, colval)
                if not colval or (isinstance(colval, str) and not colval.strip()):
                    continue
                if isinstance(colval, str):
                    
                    ## Is it an excel-ref like: 
                    #        @<sheet_name>!A1[:R10:c1].table
                    #
                    try:
                        colval = resolve_excel_ref(colval)
                    except ValueError:
                        ## Try to parse value as python-code.
                        #
                        try:
                            old_v = colval
                            colval = eval(colval)
                        except Exception:
                            pass
                        else:
                            log.info("Parsed value(%s) as python code into: %s", old_v, colval)
                datamodel.set_jsonpointer(mdl_in, colname, colval)
            
            datamodel.validate_model(mdl_in)
            experiment_pairs.append((veh_id, mdl_in))
        except Exception as ex:
            raise Exception('Invalid model for vehicle(%s): %s' % (veh_id, ex)) from ex 
        
    return experiment_pairs


def run_experiments(experiment_pairs):
    """
    Iterates `veh_df` and runs experiment_pairs.
    
    :param vehs_df:     A dataframe indexed by veh_id, and with columns *json-pointer* paths into the model
    """
    for veh_id, mdl_in in experiment_pairs:
        try:
            mdl_out = processor.run(mdl_in)
            
            add_results_as_sheet(veh_id, mdl_out)
        except Exception as ex:
            raise Exception('Experiment failed for vehicle(%s): %s' % (veh_id, ex)) from ex 


def add_results_as_sheet(veh_id, mdl_out):
    sheet = "results.%s"%veh_id
    try:
        xw.Sheet.add(sheet)
    except Exception as ex:
        log.info("Sheet(%s) already exists(%s).", sheet, ex)
        xw.Sheet(sheet).clear()
    xw.Range(sheet, 'A1', index=False).value = datamodel.resolve_jsonpointer(mdl_out, '/engine/fc_map_coeffs').to_frame().T
    xw.Range(sheet, 'A4', index=False).value = mdl_out['fitted_eng_points']
    
    

def main():        
    ## Open and run experiments
    #
    log.info('CWD: %s', os.getcwd())
    excel_fname = os.path.join(os.path.dirname(__file__), '%s.xlsm' % os.path.splitext(os.path.basename(__file__))[0])
    wb_path = os.path.abspath(excel_fname)
    xw.Workbook(wb_path)
    
    vehs_df = read_input_as_df('D2')
    log.info('%s', vehs_df)
    exp_pairs = build_models(vehs_df)
    run_experiments(exp_pairs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
Check pdcalc's function-dependencies exploration, reporting and classes .
'''
import logging
import unittest

import itertools as it

from ..pdcalc import _build_func_dependencies_graph, harvest_func, harvest_funcs_factory, _filter_common_prefixes, \
    Dependencies, DependenciesError, _validate_func_relations


def gen_all_prefix_pairs(path):
    ''' R.foo.com' --> [('R.foo.com', 'R.foo'), ('R.foo', 'R')] but outer reversed'''
    (it1, it2) = it.tee(path.split('.'))
    s1 = ''
    s2 = next(it2)
    try:
        while(True):
            s1 += next(it1)
            s2 += '.' + next(it2)
            yield (s2, s1)
            s1 += '.'
    except StopIteration:
        pass


def lstr(lst):
    return '\n'.join([str(e) for e in lst])



def funcs_fact(params, engine, dfin):
    from math import pi

    def f1(): engine['fuel_lhv'] = params['fuel'][engine['fuel']]['lhv']
    def f2(): dfin['n']     = dfin.n_norm * (engine.n_rated - engine.n_idle) + engine.n_idle
    def f3(): dfin['p']       = dfin.p_norm * engine.p_max
    def f4(): dfin['fc']      = dfin.fc_norm * engine.p_max
    def f5(): dfin['rps']     = dfin.n / 60
    def f6(): dfin['torque']  = (dfin.p * 1000) / (dfin.rps * 2 * pi)
    def f7(): dfin['pme']     = (dfin.torque * 10e-5 * 4 * pi) / (engine.capacity * 10e-6)
    def f8(): dfin['pmf']     = ((4 * pi * engine.fuel_lhv) / (engine.capacity * 10e-3)) * (dfin.fc / (3600 * dfin.rps * 2 * pi)) * 10e-5
    def f9(): dfin['cm']      = dfin.rps * 2 * engine.stroke / 1000

    return (f1, f2, f3, f4, f5, f6, f7, f8, f9)

def funcs_fact2(dfin, engine, dfout):
    def f0(): return dfin.cm + dfin.pmf + dfin.pme

    def f1(): engine['eng_map_params']      = f0()
    def f2(): dfout['n','p','fc'] = engine['eng_map_params']
    def f3(): dfout['fc_norm'] = dfout.fc / dfout.p

    return (f1, f2, f3)



class Test(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)

    def test_filter_common_prefixes(self):
        _filter_common_prefixes
        deps = ['a', 'a.b', 'b.cc', 'a.d', 'b', 'ac', 'a.c']
        res = _filter_common_prefixes(deps)
        self.assertEqual(res, ['a.b', 'a.c', 'a.d', 'ac', 'b.cc'])

        deps = ['R.dfin.hh.tt', 'R.dfin.hh.ll', 'R.dfin.hh']
        res = _filter_common_prefixes(deps)
        self.assertEqual(res, ['R.dfin.hh.ll', 'R.dfin.hh.tt'])

    def test_gen_all_prefix_pairs(self):
        path = 'R.foo.com'
        res = gen_all_prefix_pairs(path)
        self.assertEqual(list(res), [('R.foo', 'R'), ('R.foo.com', 'R.foo')])



    def testLambda(self):
        func = lambda dfin: dfin.hh['tt']

        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.tt', []), deps)

    def testLambda_successors(self):
        func = lambda dfin: dfin.hh['tt'].ss
        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.tt.ss', []), deps)
        self.assertEqual(len(deps), 1, deps)

        func = lambda dfin: dfin.hh['tt'].ss('some_arg')
        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.tt.ss', []), deps)
        self.assertEqual(len(deps), 1, deps)

        func = lambda dfin: dfin.hh['tt'].ss['oo']
        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.tt.ss.oo', []), deps)
        self.assertEqual(len(deps), 1, deps)

    def testLambda_indirectIndex(self):
        func = lambda dfin, params: dfin(params.hh['tt'])
        deps = harvest_func(func); print(deps)
        self.assertEqual(deps[0][0:2], ('R.dfin', []), deps)
        self.assertEqual(deps[1][0:2], ('R.params.hh.tt', []), deps)

    def testLambda_multiIndex(self):
        func = lambda dfin, params: dfin.hh[['tt','ll']] + params.tt
        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.ll', []), deps)
        items = [item for (item, _, _) in deps]
        self.assertEqual(items, ['R.dfin.hh.ll', 'R.dfin.hh.tt', 'R.params.tt'], deps)

    def testLambda_sliceIndex(self):
        func = lambda dfin, params: dfin.hh['tt':'ll', 'ii']
        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.ii', []), deps)
        self.assertEqual(deps[1][0:2], ('R.dfin.hh.ll', []), deps)
        self.assertEqual(deps[2][0:2], ('R.dfin.hh.tt', []), deps)
        self.assertEqual(len(deps), 3, deps)

    def testLambda_mixIndex(self):
        func = lambda dfin, params: dfin.hh['tt':'ll', 'i', params.b] + params.tt
        deps = harvest_func(func)
        items = [item for (item, _, _) in deps]
        self.assertEqual(items, ['R.dfin.hh.i', 'R.dfin.hh.ll', 'R.dfin.hh.tt', 'R.params.b', 'R.params.tt'], deps)

    @unittest.expectedFailure
    def testLambda_mixIndexSuccesor(self):
        func = lambda dfin, params: dfin.hh['tt':'ll', 'i', params.b]['g'] + params.tt
        deps = harvest_func(func)
        items = [item for (item, _, _) in deps]
        ## Cannot generate all combinations, but the last one!!
        self.assertEqual(items, ['R.dfin.hh.i.g', 'R.dfin.hh.ll.g', 'R.dfin.hh.tt.g', 'R.params.b', 'R.params.tt'], deps)


    def testFunc_sliceIndex(self):
        def func(dfin, params): dfin.hh['tt':'ll', 'ii']
        deps = harvest_func(func)
        self.assertEqual(deps[0][0:2], ('R.dfin.hh.ii', []), deps)
        self.assertEqual(deps[1][0:2], ('R.dfin.hh.ll', []), deps)
        self.assertEqual(deps[2][0:2], ('R.dfin.hh.tt', []), deps)
        self.assertEqual(len(deps), 3, deps)


    def test_harvest_funcs_factory(self):
        def func_fact(dfin, params):
            def f0(): dfin.hh['tt'].kk['ll']  = params.OO['PP'].aa
            def f1(): dfin.hh[['tt','ll', 'i']]    = params.tt
            def f2(): dfin.hh['tt':'ll', 'i']    = params.tt
            def f3(): dfin.hh['tt']

            return (f0, f1, f2, f3)

        deps = harvest_funcs_factory(func_fact)
#         self.assertEqual(deps[0][0], 'R.dfin.hh.tt.kk.ll', deps)
#         self.assertEqual(deps[0][1], 'R.params.OO.PP.aa', deps)

        self.assertEqual(deps[1][0], 'R.dfin.hh.tt', deps)
        self.assertEqual(deps[2][0], 'R.dfin.hh.ll', deps)
        self.assertEqual(deps[3][0], 'R.dfin.hh.i', deps)
        self.assertEqual(deps[1][1], ['R.params.tt'], deps)

        self.assertEqual(deps[4][0], 'R.dfin.hh.tt', deps)
        self.assertEqual(deps[5][0], 'R.dfin.hh.ll', deps)
        self.assertEqual(deps[6][0], 'R.dfin.hh.i', deps)

        self.assertEqual(deps[7][0], 'R.dfin.hh.tt', deps)

    @unittest.expectedFailure
    def test_harvest_lambas_factory(self):
        def func_fact(dfin, params):
            return [
                lambda: dfin.hh['tt'].kk['ll'] + params.OO['PP'].aa,
                lambda: dfin.hh[['tt','ll', 'i']] + params.tt,
                lambda: dfin.hh['tt':'ll', 'i'] + params.tt,
                lambda: dfin.hh['tt'],
            ]

        deps = harvest_funcs_factory(func_fact)
        self.assertEqual(deps[1][0], 'R.dfin.hh.tt.kk.ll', deps)
        self.assertEqual(deps[2][0], 'R.params.OO.PP.aa', deps) ## TODO: Why it fails?

        self.assertEqual(deps[1][0], 'R.dfin.hh.tt', deps)
        self.assertEqual(deps[2][0], 'R.dfin.hh.ll', deps)
        self.assertEqual(deps[3][0], 'R.dfin.hh.i', deps)
        self.assertEqual(deps[1][1], ['R.params.tt'], deps)

        self.assertEqual(deps[4][0], 'R.dfin.hh.tt', deps)
        self.assertEqual(deps[5][0], 'R.dfin.hh.ll', deps)
        self.assertEqual(deps[6][0], 'R.dfin.hh.i', deps)

        self.assertEqual(deps[7][0], 'R.dfin.hh.tt', deps)



    def testGatherDeps_smoke(self):
        deps = harvest_funcs_factory(funcs_fact)
        print('\n'.join([str(s) for s in deps]))


    def testGatherDepsAndBuldGraph_smoke(self):
        deps = harvest_funcs_factory(funcs_fact)
        print('\n'.join([str(s) for s in deps]))

        g = _build_func_dependencies_graph(deps)
#         print(g.edge)
#         print('topological:', '\n'.join(nx.topological_sort(g)))
#         print('topological_recusrive:', '\n'.join(nx.topological_sort_recursive(g)))
#         for line in sorted(nx.generate_edgelist(g)):
#             print(line)

    def testGatherDepsAndBuldGraph_countNodes(self):
        deps = harvest_funcs_factory(funcs_fact)
        print('\n'.join([str(s) for s in deps]))

        g = _build_func_dependencies_graph(deps)
        self.assertEqual(len(g), 20) #23 when adding.segments


    def build_web(self):
        rels = list()
        harvest_funcs_factory(funcs_fact, func_rels=rels)
        harvest_funcs_factory(funcs_fact2, func_rels=rels)
        graph = _build_func_dependencies_graph(rels)

        return graph

    def testGatherDepsAndBuldGraph_multiFuncsFacts_countNodes(self):
        import networkx as nx
        web = self.build_web()
        print("RELS:\n", lstr(web))
        print('ORDERED:\n', lstr(nx.topological_sort(web)))
        self.assertEqual(len(web), 25) #29 when adding.segments

        return web



    def test_validate_func_relations_FAIL(self):
        cases = [
            (('some.item', ('a.dep', 'a.dep.other'), None), 'prefixed with root'),
            (('R.some.item', ('a.dep', 'a.dep.other'), None), 'prefixed with root'),
            (('R.some.item', ('R.a.dep', 'a.dep.other'), None), 'prefixed with root'),
            (('some.item', ('R.a.dep', 'R.a.dep.other'), None), 'prefixed with root'),

            ((('R.some.item', ), ('R.a.dep', 'R.a.dep.other'), None), 'Bad explicit func_relations'),
            ((123, ('R.a.dep', 'R.a.dep.other'), None), 'Bad explicit func_relations'),
            (('R.some.item', ('R.a.dep', ('R.a.dep.other', )), None), 'Bad explicit func_relations'),
            (('R.some.item', (123, 'R.a.dep.other'), None), 'Bad explicit func_relations'),

#             (('R.some.item', ('R.a.dep', 'R.a.dep.other'), []), 'are DepFunc instances'),
#             ((123, ('R.a.dep', 'R.a.dep.other'), (funcs_fact, -1)), 'are DepFunc instances'),
        ]
        for (i, (rel, err)) in enumerate(cases):
            with self.assertRaisesRegex(DependenciesError, err, msg='Case(%i)'%i):
                _validate_func_relations([rel])



    def test_Dependencies_FAIL(self):
        cases = [
            ((('some.item', ), ('a.dep', 'a.dep.other'), None), 'Failed adding explicit func_relation'),
            ((123, ('a.dep', 'a.dep.other'), None), 'Failed adding explicit func_relation'),
            (('some.item', ('a.dep', ('a.dep.other', )), None), 'Failed adding explicit func_relation'),
            (('some.item', (123, 'a.dep.other'), None), 'Failed adding explicit func_relation'),
        ]

        deps = Dependencies()

        for (i, (rel, err)) in enumerate(cases):
            with self.assertRaisesRegex(DependenciesError, err, msg='Case(%i)'%i):
                deps.add_func_rel(*rel)

    def test_Dependencies_GOOD(self):
        cases = [
            ('some.item', ('a.dep', 'a.dep.other'), None),
            ('R.some.item', ('a.dep', 'a.dep.other'), None),
            ('R.some.item', ('R.a.dep', 'a.dep.other'), None),
            ('some.item', ('R.a.dep', 'R.a.dep.other'), None),

            ('some.item', ('a.dep', 'a.dep.other'), None),
            ('some.item', ('a.dep', 'a.dep.other'), funcs_fact),
            ('some.item', ('a.dep', 'a.dep.other'), (funcs_fact, 1)),
        ]

        deps = Dependencies()

        for rel in cases:
            deps.add_func_rel(*rel)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
The core calculations required for transforming the Input-datamodel to the Output one.

Uses *pandalon*'s automatic dependency extraction from calculation functions.
"""
import logging

import numpy as np
import pandas as pd
import lmfit 

from . import pdcalc
from . import datamodel
from collections import OrderedDict


log = logging.getLogger(__name__)


def run(mdl, opts=None):
    """
    :param mdl: the datamodel to process, all params and data
    :param map opts: flags controlling non-functional aspects of the process (ie error-handling and logging, gui, etc)
    """

    datamodel.ensure_modelpath_Series(mdl, '/engine')
    #datamodel.ensure_modelpath_Series(mdl, '/params')
    datamodel.ensure_modelpath_DataFrame(mdl, '/measured_eng_points')

    params              = mdl['params']
    engine              = mdl['engine']
    measured_eng_points = mdl['measured_eng_points']

    ## Identify quantities necessary for the FITTING, 
    #    and calculate them.
    outcomes = ('eng_points.cm', 'eng_points.pme', 'eng_points.pmf', 'engine.fuel_lhv')
    pdcalc.execute_funcs_factory(eng_points_2_std_map, outcomes, params, engine, measured_eng_points)

    ## FIT
    #
    coeffs = datamodel.resolve_jsonpointer(mdl, '/params/fitting/coeffs')
    coeffs = [lmfit.parameter.Parameter(name, **kws) for (name, kws) in coeffs.items()]
    fitted_coeffs = fit_engine_map(measured_eng_points, 
                datamodel.resolve_jsonpointer(mdl, '/params/fitting/is_robust'),
                coeffs,
                datamodel.resolve_jsonpointer(mdl, '/params/fitting/fit_options'),
            )
    engine['fc_map_coeffs'] = fitted_coeffs

    fitted_eng_points   = reconstruct_eng_points_fitted(engine, fitted_coeffs, measured_eng_points)
    std_to_norm_map(engine, fitted_eng_points)

    if datamodel.resolve_jsonpointer(mdl, '/params/plot_maps'):
        mesh_eng_points     = generate_mesh_eng_points_fitted(measured_eng_points, fitted_coeffs, measured_eng_points)
        columns = ['pmf', 'cm', 'pme']
        plot_map(measured_eng_points, mesh_eng_points, columns)
        
        ## Flatten 2D-vectors to make a DataFrame
        #
        mesh_eng_points = {col: vec.flatten() for (col, vec) in mesh_eng_points.items()}
        mesh_eng_points = pd.DataFrame(mesh_eng_points)
        ## Fill calced columns.
        #
        std_to_norm_map(engine, mesh_eng_points)
        
        mdl['mesh_eng_points'] = pd.DataFrame(mesh_eng_points)
        

    mdl['measured_eng_points'] = measured_eng_points
    mdl['fitted_eng_points'] = pd.DataFrame(fitted_eng_points)

    #datamodel.validate_model(mdl, additional_properties=False) TODO: Make OUT-MODEL pass validation. 
    
    return mdl


def eng_points_2_std_map(params, engine, eng_points):
    """
    A factory of the calculation functions for reaching to the data necessary for the Fitting.

    The order of the functions below not important, and the actual order of execution 
    is calculated from their dependencies, based on the data-frame's column access.
    """
    from math import pi
    
    funcs = [
        lambda: engine.__setitem__('fuel_lhv',           params['fuel'][engine.fuel]['lhv']),
        lambda: eng_points.__setitem__('n',      eng_points.n_norm * (engine.n_rated - engine.n_idle) + engine.n_idle),
        lambda: eng_points.__setitem__('p',      eng_points.p_norm * engine.p_max),
        lambda: eng_points.__setitem__('fc',     eng_points.fc_norm * engine.p_max),
        lambda: eng_points.__setitem__('rps',    eng_points.n / 60),
        #lambda: eng_points.__setitem__('rps',     eng_points.cm * 1000 / (2 * engine.stroke)),
        lambda: eng_points.__setitem__('torque', (eng_points.p * 1000) / (eng_points.rps * 2 * pi)),
        lambda: eng_points.__setitem__('pme',    (eng_points.torque * 10e-5 * 4 * pi) / (engine.capacity * 10e-6)),
        lambda: eng_points.__setitem__('pmf',    ((4 * pi * engine.fuel_lhv) / (engine.capacity * 10e-6)) * (eng_points.fc / (3600 * eng_points.rps * 2 * pi)) * 10e-5),
        lambda: eng_points.__setitem__('cm',     eng_points.rps * 2 * engine.stroke / 1000),
    ]

    return funcs

def std_to_norm_map(engine, eng_points):
    from math import pi

    eng_points['rps']       = eng_points.cm * 1000 / (2 * engine.stroke)
    eng_points['n']         = eng_points.rps * 60
    eng_points['n_norm']    = eng_points.n / (engine.n_rated - engine.n_idle) + engine.n_idle
    
    eng_points['torque']    = eng_points.pme * (engine.capacity * 10e-3) / (4 * pi * 10e-5)
    eng_points['p']         = eng_points.torque * (eng_points.rps * 2 * pi) / 1000

    eng_points['fc']        = (eng_points.pmf * (engine.capacity * 10e-2) * (3600 * eng_points.rps * 2 * pi)) / (4 * pi * engine.fuel_lhv * 10e-5)
    eng_points['fc_norm']   = eng_points.fc / engine.p_max

    eng_points['p_norm']    = eng_points.p / engine.p_max



def engine_map_modelfunc(coeff_values, X):
    """
    The function that models the engine-map.
    """
    
    a = coeff_values['a']
    b = coeff_values['b']
    c = coeff_values['c']
    a2 = coeff_values['a2']
    b2 = coeff_values['b2']
    loss0 = coeff_values['loss0']
    loss2 = coeff_values['loss2']

    pmf = X['pmf']
    cm = X['cm']
    
    pme = (a + b*cm + c*cm**2)*pmf + (a2 + b2*cm)*pmf**2 + loss0 + loss2*cm**2
    
    return pme


def fit_engine_map(df, is_robust, coeffs, fit_options):
    assert len({'cm', 'pme', 'pmf'} - set(df.columns)) == 0, \
            "Missing fit-columns: %s" % {'cm', 'pme', 'pmf'} - set(df.columns)
    assert not np.any(np.isnan(df['pmf'])), \
            "Cannot fit with NaNs in `pmf` data! \n%s" % np.any(np.isnan(df['pmf']), axis=1)
    assert not np.any(np.isnan(df['cm'])), \
            "Cannot fit with NaNs in `cm` data! \n%s" % np.any(np.isnan(df['cm']), axis=1)

    residualfunc_args   = (engine_map_modelfunc, df, df['pme'])
    residualfunc_kws    = dict(is_robust=is_robust)
    minimizer = lmfit.minimize(_robust_residualfunc, coeffs, 
                args=residualfunc_args, 
                kws=residualfunc_kws,
                **fit_options)
    res_df = pd.Series(minimizer.params.valuesdict())

    return res_df


def reconstruct_eng_points_fitted(engine, fitted_coeffs, eng_points):
    pme = engine_map_modelfunc(fitted_coeffs, eng_points)

    fitted_eng_points = pd.DataFrame.from_items(zip(['pmf', 'cm', 'pme'], (eng_points.pmf, eng_points.cm, pme)))

    return fitted_eng_points

def generate_mesh_eng_points_fitted(engine, fitted_coeffs, eng_points):
    ## Construct X.
    #
    dmin = eng_points.loc[:, ['pmf', 'cm']].min(axis=0)
    dmax = eng_points.loc[:, ['pmf', 'cm']].max(axis=0)
    drng = (dmax - dmin)
    dmin -= 0.05 * drng
    dmax += 0.10 * drng
    dstp = (dmax - dmin) / 40

    X1, X2 = np.mgrid[dmin[0]:dmax[0]:dstp[0], dmin[1]:dmax[1]:dstp[1]]

    X = {'pmf': X1, 'cm': X2}
    Y = engine_map_modelfunc(fitted_coeffs, X)


    mesh_eng_points = OrderedDict(zip(['pmf', 'cm', 'pme'], (X1, X2, Y)))

    return mesh_eng_points


def plot_map(dfin, fitted_eng_points, columns):
    (X1, X2, Y) = [fitted_eng_points[col] for col in columns]

    x1min = X1.min(); x1max = X1.max();
    x2min = X2.min(); x2max = X2.max();
    extent=(x1min, x1max, x2min, x2max)
    levels = np.arange(Y.min(), Y.max(), (Y.max() - Y.min()) / 10.0)

#     import matplotlib
#     matplotlib.use('WebAgg')
    from matplotlib import pyplot as plt

    plt.plot(dfin[columns[0]], dfin[columns[1]], '.c')

    cntr = plt.contourf(X1, X2, Y, cmap=plt.cm.get_cmap(plt.cm.copper, len(levels)-1), extent=extent)  # @UndefinedVariable
    colorbar = plt.colorbar(cntr)
    colorbar.set_label(columns[2], color='blue')

    ax = plt.gca()
    ax.set_title('Fitted normalized engine_map')
    ax.set_aspect('auto'); ax.set_adjustable('box-forced')
    ax.set_xlabel(columns[0], color='red'); ax.set_ylabel(columns[1], color='green')

    plt.show()


def proc_vehicle(dfin, datamodel):

    ## Filter values
    #
    nrows       = len(dfin)
    dfin = dfin[(dfin.pme > -1.0) & (dfin.pme < 25.0)]
    log.warning('Filtered %s out of %s rows for BAD  pme.', nrows-len(dfin), nrows)

    return dfin



def _robust_residualfunc(coeffs, modelfunc, X, YData, is_robust=False, robust_prcntile=None):
    """
    A non-linear iteratively-reweighted least-squares (IRLS) residual function (objective-function) 
    that robustly fits ``YData = modelfunc(X)``.

    This method applies weights on each iteration so as to downscale any outliers and high-leverage data-points
    based on the 'bisquare' standardized adjusted residuals:[#]_
    
    .. math::
    
            \frac{r}{K \times \hat{\sigma} \times \sqrt{1 - h}}
    
    where:

    :math:`r` (vector)
        the residuals :math:`\hat{y} - y`
        
    :math:`K` (scalar)
        the *robust percentile* tuning constant used on each iteration to filter-out
        adjusted-standardized-weights above 1, expressed as the Bisquare M-estimator efficiency 
        under Gaussian model. 
        
    :math:`\hat{\sigma}` (scalar)
        the robust estimate of the *standard deviation* of the residuals
        based on MAD[#]_ like this: :math:`\hat{\sigma}=1.4826\times\operatorname{MAD}`
        
    :math:`h` : (vector)
        the *hat vector*, the diagonal of the *hat matrix*,[#]_
        which is used to reduce the weight of high-leverage data points
        that are having a large effect on the least-squares fit.


    :param modelfunc:             The modeling function that accepts the dict of coeffs
    :param nparray X:             
    :param nparray YData:         measured-data points
    :param boolean is_robust:     Whether to deleverage outlier YData.
    :param float robust_prcntile: The `K` percentile of the MAD, 
                             [default: 4.68, filters-out approximately 5% of the residuals as outliers]

    .. Seealso::
        curve_fit, leastsq


    .. [#] http://www.mathworks.com/help/stats/robustfit.html
    .. [#] https://en.wikipedia.org/wiki/Median_absolute_deviation
    .. [#] https://en.wikipedia.org/wiki/Hat_matrix
    """
    YFitted = engine_map_modelfunc(coeffs.valuesdict(), X)
    Residual    = YFitted - YData

    ## Robust:
    ## Deleverage and standardize absolute-residuals based on a robust-MAD.
    ##
    if is_robust:
        R_abs       = Residual.abs()
        if not robust_prcntile:
            robust_prcntile = 4.685     ##  Bisquare M-estimator with 95% efficiency under the Gaussian model.
        R_deleved   = R_abs / (robust_prcntile * 1.4826 * np.median(R_abs))  ## * sqrt(1 - hat_vector))
        ## Calc the robust bisquared-residuals excluding outliers.
        R_weights   = (R_deleved < 1) * (1 - R_deleved**2)**2

        Residual = R_weights * Residual
        
    return Residual





#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
Check cmdline parsing and model building.
'''

from argparse import ArgumentTypeError
import argparse
from collections import OrderedDict
import functools
import glob
import io
import logging
import os
import shutil
import sys
import tempfile
import unittest

from ..__main__ import (
    build_args_parser, validate_file_opts, parse_key_value_pair,
    parse_many_file_args, assemble_model,
    FileSpec, main, store_model_parts
)
from ..__main__ import parse_column_specifier

from ..datamodel import json_dumps, base_model, validate_model
from .redirect import redirected # @UnresolvedImport


def _init_logging(loglevel):
    logging.basicConfig(level=loglevel)
    logging.getLogger().setLevel(level=loglevel)

_init_logging(logging.INFO)
log = logging.getLogger(__name__)


def from_my_path(*parts):
    return os.path.join(os.path.dirname(__file__), *parts)

def copy_data_files_to_cwd():
    copy_paths = ['*.xlsx', '*.csv']
    for path in copy_paths:
        for f in glob.glob(from_my_path(path)):
            shutil.copy(f, '.')


class TestFuncs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()
        os.chdir(cls.temp_dir.name)


    def setUp(self):
        self._test_fnames = ('~temp.csv', '~strange.ext')
        for tfn in self._test_fnames:
            if (not os.path.exists(tfn)):
                tf = open(tfn, "w")
                tf.close()
        copy_data_files_to_cwd()
            
        _exit_code = None
        _exit_msg = None

        self._failColumnsMsg = 'Not a COLUMN_SPEC syntax'
        self._failColumns = ['', '(dd)', 'quant((bad units))', 'quant --> ((bad units))', 'quant (complex ( units ))',
                'quant (open units']
        self._goodColumns = [
            ('quant',                   ('quant', None)),
            ('spaced quant',            ('spaced quant', None)),
            ('quant (units)',           ('quant', 'units')),
            ('quant (spaced units)',    ('quant', 'spaced units')),
            ('Pnorm (kJ / sec)',        ('Pnorm', 'kJ / sec')),

            ('quant [units]',           ('quant', 'units')),
            ('quant [spaced units]',    ('quant', 'spaced units')),
            ('Pnorm [kJ / sec]',        ('Pnorm', 'kJ / sec')),
            ('Pnorm [kJ/(sec)]',        ('Pnorm', 'kJ/(sec)')),
            ('Pnorm [(kJ*2)/(sec)]',    ('Pnorm', '(kJ*2)/(sec)')),
        ]

        self._failKVPairsMsg = 'Not a KEY=VALUE syntax'
        self._failKVPairs = ['no_value', 'spa ced', 'spaced key = key', '3number=val',
            '_']
        self._goodKVPairs = {'k=v':'v', 'key=value':'value', 'Num_3_key = 3 ':'3', '_key=hidden_key':'hidden_key',
            'k = spaced value ':'spaced value', 'miss_value=':'',
            'k+=123':123, 'k*=12.3':12.3, 'k?=1':True, 'k?=true':True, 'k?=False':False, 'k?=on':True,
            'k := 3':3, 'k := 3.14':3.14, 'k := ["a", 3.14]':['a', 3.14], 'k:={"a":3.14, "2":[1, "1"]}':{'a':3.14, '2':[1, "1"]},
            'k @= 3':3, 'k @= 3.14':3.14, 'k @= ["a", 3.14]':['a', 3.14], 'k@={\'a\':3.14, "2":[1, "1"]}':{'a':3.14, '2':[1, "1"]},
            'path/key += 0': 0, '/rootpath/+=1':1
        }

        self._failFormatsMsg = 'invalid choice:'
        self._failFormats = ['file_frmt=', 'file_frmt=BAD', 'file_frmt=CSV BAD']
        self._goodFormats = ['file_frmt=AUTO', 'file_frmt=CSV', 'file_frmt=XLS']


    def get_args_parser(self):
        def _exit(p, status=0, message=None):
            if message:
                self._exit_msg = message
            self._exit_code = status

#         def _error(p, message):
#             self._exit_msg = message
#             self._exit_code = 2
        parser = build_args_parser('test', 'x.x.x', 'DESC', 'EPILOG')
#         parser.exit = _exit
#         parser.error = _error

        return parser


    def check_exits(self, cmdline, exit_code):
        mystdout = io.StringIO()
        mystderr = io.StringIO()
        opts = None  # @UnusedVariable
        with redirected(stdout=mystdout, stderr=mystderr):
            with self.assertRaises(SystemExit) as cm:
                parser = self.get_args_parser()
                opts = parser.parse_args(cmdline)

            if (cm.exception.code != exit_code):
                d = OrderedDict({
                     'CMDLNE': cmdline,
                     'STDOUT': mystdout.getvalue(),
                     'STDERR': mystderr.getvalue(),
                     'OPTS': opts,
                 })
                raise AssertionError(d)
        return opts, mystdout, mystderr



    def check_ok(self, cmdline, **kw):
        import io

        mystdout = io.StringIO()
        mystderr = io.StringIO()
        opts = None  # @UnusedVariable
        with redirected(stdout=mystdout, stderr=mystderr, **kw):
            try:
                parser = self.get_args_parser()
                opts = parser.parse_args(cmdline)
            except SystemExit as ex:
                d = OrderedDict({
                     'EX': ex,
                     'CMDLNE': cmdline,
                     'STDOUT': mystdout.getvalue(),
                     'STDERR': mystderr.getvalue(),
                     'OPTS': opts,
                 })
                raise AssertionError(d)

        return opts, mystdout, mystderr


    def checkParseOpt_fail(self, cmdline, bads, badMsg, exit_code = 2):
        for opt in bads:
            (opts, mystdout, mystderr) = self.check_exits(cmdline.split() + [opt], exit_code)

            self.assertIsNone(opts)
            self.assertFalse(mystdout.getvalue(), mystdout.getvalue())
            self.assertIn(badMsg, mystderr.getvalue().splitlines()[-1], mystderr.getvalue())

        ## Test multiple bads in 1 line
        self.check_exits(cmdline.split() + bads, exit_code)


    def checkParseOpt_good(self, cmdline, goods, testNArgs=True):
        for opt in goods:
            (opts, mystdout, mystderr) = self.check_ok(cmdline.split() + [opt])

            self.assertIsNotNone(opts)
            self.assertFalse(mystderr.getvalue(), mystderr.getvalue())
            self.assertFalse(mystdout.getvalue(), mystdout.getvalue())

        ## Test multiple goods in 1 line
        if testNArgs:
            self.check_ok(cmdline.split() + list(goods))


    def checkColumnFormat_bad(self, frmts, ex_msg):
        for frmt in frmts:
            with self.assertRaisesRegex(ArgumentTypeError, ex_msg, msg=frmt):
                parse_column_specifier(frmt)

    def checkColumnFormat_good(self, frmt_cases):
        for (frmt, (quant, units)) in frmt_cases:
            res = parse_column_specifier(frmt)
            self.assertEqual(res['name'], quant, (frmt, res))
            self.assertEqual(res['units'], units, (frmt, res))


    def testHelpMsg(self):

        cmdline = '--help'
        exit_code = 0

        (opts, mystdout, mystderr) = self.check_exits(cmdline.split(), exit_code)

        self.assertIsNone(opts)
        self.assertFalse(mystderr.getvalue(), mystderr.getvalue())
        ## On errors, help-msg is not printed.
        self.assertIn("DESC", mystdout.getvalue(), mystdout.getvalue())  # @UndefinedVariable


    def testModelOverrides_fail(self):
        self.checkParseOpt_fail('-I funcs_map -m', [''] + self._failKVPairs, self._failKVPairsMsg)
    def testModelOverrides_good(self):
        self.checkParseOpt_good('-I funcs_map -m', self._goodKVPairs.keys())


    def testColumnNames_fail(self):
        self.checkColumnFormat_bad(self._failColumns, ex_msg = 'Not a COLUMN_SPEC syntax')
    def testColumnNames_good(self):
        self.checkColumnFormat_good(self._goodColumns)

    def testColumnNamesInParser_fail(self):
        self.checkParseOpt_fail('-c', self._failColumns, self._failColumnsMsg)
    def testColumnNamesInParser_good(self):
        cols = [frmt for (frmt, _) in self._goodColumns]
        self.checkParseOpt_good('-I ~temp.csv -c', cols)

    def testColumnRenamesInParser_fail(self):
        self.checkParseOpt_fail('-r', self._failColumns, self._failColumnsMsg)
    def testColumnRenamesInParser_good(self):
        cols = [frmt for (frmt, _) in self._goodColumns]
        self.checkParseOpt_good('-I funcs_map -r', cols)


    def testKVPairs_fail(self):
        self._failKVPairs
        for opt in self._failKVPairs:
            with self.assertRaises(argparse.ArgumentTypeError, msg=opt):
                parse_key_value_pair(opt)
    def testKVPairs_good(self):
        self._goodKVPairs
        for (arg, val) in self._goodKVPairs.items():
            (_, v) = parse_key_value_pair(arg)
            self.assertEqual(v, val, arg)

    def testFileSpec_fail(self):
        test_fnames = self._test_fnames
        cases = [
            [['']],
            [['missing.file']],
            [[test_fnames[0], 'file_frmt=BAD']],
            [[test_fnames[0], 'model_path=rel_path']],
            [[test_fnames[0], '2_bad=key']],
            [[test_fnames[1]]], # missing file_frmt
        ]
        for many_file_args in cases:
            with self.assertRaises(argparse.ArgumentTypeError, msg=many_file_args):
                parse_many_file_args(many_file_args, 'r', None)
            with self.assertRaises(argparse.ArgumentTypeError, msg=many_file_args):
                parse_many_file_args(many_file_args, 'r')

        all_cases = functools.reduce(lambda x, y: x+y, cases)
        with self.assertRaises(argparse.ArgumentTypeError, msg=all_cases):
            parse_many_file_args(all_cases, 'r')
        with self.assertRaises(argparse.ArgumentTypeError, msg=all_cases):
            parse_many_file_args(all_cases, 'r')

    def testFileSpec_good(self):
        test_fnames = self._test_fnames
        cases = [
            (('r', 'w'), [[test_fnames[0]]]),
            (('r', 'w'), [[test_fnames[0], 'file_frmt=AUTO']]),
            (('r', 'w'), [[test_fnames[0], 'model_path=']]),
            (('r', 'w'), [[test_fnames[0], 'file_frmt=CSV', 'model_path=/gjhgj']]),
            (('r', 'w'), [[test_fnames[0], 'some=other', 'keys+=4', 'fun:=[1, {"a":2}]']]),
            (('r', 'w'), [[test_fnames[1], 'file_frmt=CSV']]),
            (('r', 'w'), [['+', ]]),

            (('r', 'w'), [['-', ]]),
            (('r', 'w'), [['+', ]]),
            (('r', 'w'), [['-', 'file_frmt=JSON']]),
            (('r', 'w'), [['+', 'file_frmt=JSON']]),
            (('r', 'w'), [['-', 'file_frmt=CSV', 'model_path=/gjhgj']]),
            (('r', 'w'), [['+', 'file_frmt=CSV', 'model_path=/gjhgj']]),
        ]
        for (open_modes, many_file_args) in cases:
            for open_mode in open_modes:
                parse_many_file_args(many_file_args, open_mode, None)

        argparse.ArgumentTypeError(parse_many_file_args, functools.reduce(lambda x, y: x+y, cases), 'r')


    def testNumOfFileOpts_fail(self):
        cases = [
               {'I':None, 'icolumns':[1,2], 'irenames':None},
               {'I':[1], 'icolumns':[1,2], 'irenames':None},
               {'I':[1,2], 'icolumns':[1,2,3], 'irenames':None},

               {'I':None, 'irenames':[1,2], 'icolumns':None},
               {'I':[1], 'irenames':[1,2], 'icolumns':None},
               {'I':[1,2, 3], 'irenames':[1,2], 'icolumns':None},

               {'I':None, 'irenames':[1,2], 'icolumns':[1,2]},
               {'I':None, 'irenames':[1,2], 'icolumns':[1]},
               {'I':[1], 'irenames':[1,2], 'icolumns':[1]},
               {'I':[1], 'irenames':[1,2], 'icolumns':[1,2,3]},
               {'I':[1,2], 'irenames':[1,2,3], 'icolumns':None},
               {'I':[1,2,3], 'irenames':[1,2], 'icolumns':[1,2,3,4]},
               {'I':[1,2,3], 'irenames':[1], 'icolumns':[1,2,3,4]},
       ]
        for opts in cases:
            opts = argparse.Namespace(**opts)
            with self.assertRaises(argparse.ArgumentTypeError, msg=opts):
                validate_file_opts(opts)

    def testNumOfFileOpts_good(self):
        cases = [
               {'I':[1], 'icolumns':[1], 'irenames':None},
               {'I':[1,2], 'icolumns':None, 'irenames':[1,2]},
               {'I':[1,2], 'icolumns':[1,2], 'irenames':[1,2]},
               {'I':[1,2], 'icolumns':[1,2], 'irenames':[1]},
               {'I':[1,2], 'icolumns':[1], 'irenames':[1]},
       ]
        for opts in cases:
            opts = argparse.Namespace(**opts)
            validate_file_opts(opts)

    def testSmoke_BuildModel_model_overrideParse_n_print(self):
        import pandas as pd

        fname = from_my_path('test_table.csv')
        opts = {'m':[[('fuel','diesel')]] }
        filespecs = [
            FileSpec(pd.read_csv, fname, open(fname, 'r'), 'CSV', '/measured_eng_points', None, None, {})
        ]
        opts = argparse.Namespace(**opts)
        mdl = assemble_model(filespecs, opts.m)

        print(json_dumps(mdl))


    def testBuildModel_validate(self):
        import pandas as pd

        fname = from_my_path('test_table.csv')
        model_overrides = [[('fuel','diesel')]]
        filespecs = [
            FileSpec(pd.read_csv, fname, open(fname, 'r'), 'CSV', '/measured_eng_points', None, None, {})
        ]
        mdl = assemble_model(filespecs, model_overrides)
        validate_model(mdl)


    def testWriteModelparts_emptyModel(self):
        mystdout = io.StringIO()
        mdl = {}
        filespecs = [
            FileSpec('write_csv', '<mystream>', mystdout, 'CSV', '', None, None, {})
        ]

        mystdout = io.StringIO()
        store_model_parts(mdl, filespecs)
        self.assertEqual('', mystdout.getvalue(), mystdout.getvalue())

    def testWriteModelparts_baseModel(self):
        mystdout = io.StringIO()
        mdl = base_model()
        filespecs = [
            FileSpec('write_csv', '<mystream>', mystdout, 'CSV', '', None, None, {})
        ]
        store_model_parts(mdl, filespecs)
        self.assertEqual(json_dumps(mdl), mystdout.getvalue(), mystdout.getvalue())

    def testWriteModelparts_alteredModel(self):
        mystdout = io.StringIO()
        mdl = base_model()
        k = 'test_key'
        v = 'TEST_value'
        mdl['engine'][k] = v
        filespecs = [
            FileSpec('write_csv', '<mystream>', mystdout, 'CSV', '', None, None, {})
        ]
        store_model_parts(mdl, filespecs)
        self.assertIn(k, mystdout.getvalue(), mystdout.getvalue())
        self.assertIn(v, mystdout.getvalue(), mystdout.getvalue())

    #def testWriteModelparts_dataframe(self):

    @classmethod
    def tearDownClass(cls):
        try:
            cls.temp_dir.cleanup()
        except Exception:
            log.warning('Minor failure while cleaning up!', exc_info=True)


class TestMain(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()
        os.chdir(cls.temp_dir.name)

    def setUp(self):
        self.held, sys.stdout = sys.stdout, io.StringIO()
        copy_data_files_to_cwd()
    
    def test_run_main_stdout1(self):
        main('''-I FuelFit.xlsx  sheetname+=0 header@=None names:=["n","p","fc"]
                -I engine.csv file_frmt=SERIES model_path=/engine header@=None
                -m /engine/fuel=petrol
                -O - model_path= -v -d'''.split())
        self.assertIn('n', sys.stdout.getvalue().strip(), 0)

    def test_run_main_stdout2(self):
        main('''-I FuelFit.xlsx  sheetname+=0 header@=None names:=["n","p","fc"]
                -I engine.csv file_frmt=SERIES model_path=/engine header@=None
                -m /engine/fuel=petrol
                -O - model_path=/fitted_eng_points  index?=false -v -d'''.split())
        self.assertIn('n', sys.stdout.getvalue().strip(), 0)

    def test_run_main_fileout(self):
        out_fname = '~t.json'
        main('''-I FuelFit.xlsx  sheetname+=0 header@=None names:=["n","p","fc"]
                -I engine.csv file_frmt=SERIES model_path=/engine header@=None
                -m /engine/fuel=petrol
                -O ~t1.csv model_path=/measured_eng_points index?=false
                -O ~t2.csv model_path=/fitted_eng_points index?=false
                -O {} model_path=
                -m /params/plot_maps?=False
                '''.format(out_fname).split())
        with open(out_fname, 'r') as fp:
            txt = fp.read()
        self.assertIn('n', txt.strip())
        self.assertIn('n', txt.strip())

    def test_run_main_stdout3(self):
        main('''
            -vd
            -I FuelFit_real.csv header+=0 names@='n_norm,p_norm,fc_norm'.split(',')
            -I engine.csv file_frmt=SERIES model_path=/engine header@=None
            -m /engine/fuel=petrol'''.split())
        self.assertEqual(len(sys.stdout.getvalue().strip()), 0)

    def test_run_main_stdout4(self):
        main('''-vd
            -I FuelFit_real.csv header+=0
              --irenames n_norm _ fc_norm
            -I engine.csv file_frmt=SERIES model_path=/engine header@=None
              --irenames
            -m /engine/fuel=petrol
        '''.split())


    def test_run_main_stdout5_icolumns(self):
        main('''-vd
            -I FuelFit_real.csv header+=0
              --irenames n_norm _ fc_norm
            -I engine.csv file_frmt=SERIES model_path=/engine header@=None
              --irenames
            -m /engine/fuel=petrol
            -O - model_path=/fitted_eng_points index?=false
            -O - model_path=/engine/fc_map_coeffs
            -m /params/plot_maps@=False
        '''.split())



    def tearDown(self):
        self.held, sys.stdout = sys.stdout, self.held
        print(self.held.getvalue())

    @classmethod
    def tearDownClass(cls):
        try:
            cls.temp_dir.cleanup()
        except Exception:
            log.warning('Minor failure while cleaning up!', exc_info=True)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''A best-effort attempt to build computation dependency-graphs from method with dict-like objects (such as pandas),
inspired by XForms: http://lib.tkk.fi/Diss/2007/isbn9789512285662/article3.pdf

.. seealso::
    Dependencies
    @calculation
    @calculation_factory [TODO]
    execute_funcs_map()
'''
from collections import OrderedDict, defaultdict
from collections.abc import Mapping, Iterable
import inspect
import logging
import re

from networkx.exception import NetworkXError

import itertools as it
import networkx as nx
import pandas as pd


from .mymock import MagicMock

DEBUG= False

_root_name = 'R'
_root_len = len(_root_name)+1
log = logging.getLogger(__name__)

class DependenciesError(Exception):
    def __init__(self, msg, item=None):
        super(Exception, self).__init__(msg)
        self.item = item

def _make_mock(*args, **kwargs):
    return MagicMock(*args, **kwargs)


def harvest_funcs_factory(funcs_factory, root=None, func_rels=None):
    new_func_rels = []

    ## Wrap and invoke funcs_factory with "rooted" mockups as args
    #    to collect mock_calls.
    #
    funcs_factory = _wrap_funcs_factory(funcs_factory)
    (root, mocks) = funcs_factory.mockup_func_args(root=root)

    cfuncs = funcs_factory(*mocks) ## The cfuncs are now wrapped children.

    ## Harvest cfunc deps as a list of 3-tuple (item, deps, funx)
    #    by inspecting root after each cfunc-call.
    #
    for cfunc in cfuncs:
        root.reset_mock()
        tmp = cfunc()
        try: tmp += 2   ## Force dependencies from return values despite compiler-optimizations.
        except:
            pass
        _harvest_mock_calls(root.mock_calls, cfunc, new_func_rels)
    funcs_factory.reset()

    _validate_func_relations(new_func_rels)

    if func_rels is not None:
        func_rels.extend(new_func_rels)

    return new_func_rels

def harvest_func(func, root=None, func_rels=None):
    new_func_rels = []

    func = _wrap_standalone_func(func)
    (root, mocks) = func.mockup_func_args(root=root)

    tmp = func(*mocks)
    try: tmp += 2   ## Force dependencies from return values despite compiler-optimizations.
    except:
        pass
    func.reset()
    _harvest_mock_calls(root.mock_calls, func, new_func_rels)

    _validate_func_relations(new_func_rels)

    if func_rels is not None:
        func_rels.extend(new_func_rels)

    return new_func_rels


def _harvest_mock_calls(mock_calls, func, func_rels):
    ## A map from 'pure.dot.paths --> call.__paths__
    #  filled-in and consumed )mostly) by _harvest_mock_call().
    deps_set = OrderedDict()

    #last_path = None Not needed!
    for call in mock_calls:
        last_path = _harvest_mock_call(call, func, deps_set, func_rels)

    ## Any remaining deps came from a last not-assignment (a statement) in func.
    #  Add them as non-dependent items.
    #
    if deps_set:
        deps_set[last_path] = None  ## We don't care about call dep-subprefixes(the value) anymore.
#         if ret is None:
#             item = last_path
#         else:
#             item = _parse_mock_arg(ret)
        for dep in _filter_common_prefixes(deps_set.keys()):
            _append_func_relation(dep, [], func, func_rels)

def _parse_mock_arg(mock):
    mpath = _parse_mock_str(mock)[0]
    return (_strip_magic_tail(mpath), mpath)

def _harvest_mock_call(mock_call, func, deps_set, func_rels):
    '''Adds a 2-tuple (indep, [deps]) into indeps with all deps collected so far when it visits a __setartr__. '''

    def parse_mock_path(mock):
        mpath = _parse_mock_str(mock)[0]
        try:
            ## Hack to consolidate 'dot.__getitem__.com' --> fot.Xt.com attributes.
            #  Just search if previous call is subprefix of this one.
            prev_path = next(reversed(deps_set))
            prev_call = deps_set[prev_path]
            if (prev_call+'()' == call[:len(prev_call)+2]):
                mpath = prev_path + mpath[len(prev_call)+_root_len+2:] # 4 = R.()
        except (KeyError, StopIteration):
            pass
        return _strip_magic_tail(mpath)

    (call, args, kw) = mock_call

    deps_set.update((_parse_mock_arg(arg) for arg in args        if isinstance(arg, MagicMock)))
    deps_set.update((_parse_mock_arg(arg) for arg in kw.values() if isinstance(arg, MagicMock)))


    path = parse_mock_path(mock_call.parent)

    tail = call.split('.')[-1]
    if (tail == '__getitem__'):
        for item in _harvest_indexing(args[0]):
            new_path = '%s.%s'%(path, item)
            deps_set[new_path] = call
    elif (tail == '__setitem__'):
        deps = list(deps_set.keys())
        deps_set.clear()

        for item in _harvest_indexing(args[0]):
            new_path ='%s.%s'%(path, item)
            _append_func_relation(new_path, deps, func, func_rels)

    return path

def _append_func_relation(item, deps, func, func_rels):
    func_rels.append((item, deps, func))

def _harvest_indexing(index):
    '''Harvest any strings, slices, etc, assuming to be DF's indices. '''

    if isinstance(index, slice):
        deps = _harvest_indexing(index.start) + _harvest_indexing(index.stop) + _harvest_indexing(index.step)
    elif isinstance(index, str):
        deps = [index]
    elif isinstance(index, Iterable):
        deps = [ii for i in index for ii in _harvest_indexing(i)]
    else:
        deps = []
    return deps


def _strip_magic_tail(path):
    '''    some.path___with_.__magics__ --> some.path '''
    pa_th = path.split('.')
    while (pa_th[-1].startswith('__')):
        del pa_th[-1]
        if (not pa_th):
            return []
    path = '.'.join(pa_th)
    return path[:-2] if path.endswith('()') else path


_mock_id_regex = re.compile(r"name='([^']+)' id='(\d+)'")
def _parse_mock_str(m):
    return _mock_id_regex.search(m.__repr__()).groups()

def _validate_func_relations(func_rels):
    try:
        bad_items = [item for (item, _, _) in func_rels if not item.startswith(_root_name)]
        if bad_items:
            raise DependenciesError("Not all dependency-items are prefixed with root(%s): %s"%(_root_name, bad_items), bad_items)

        bad_deps = [dep for (_, deps, _) in func_rels for dep in deps if not dep.startswith(_root_name)]
        if bad_deps:
            raise DependenciesError("Not all dependency-data are prefixed with root(%s): %s"%(_root_name, bad_deps), bad_deps)

        ## NO, check only for algo-generated errors.
        # bad_funcs = [func for (_, _, func) in func_rels if not isinstance(func, _DepFunc)]
        # if bad_funcs:
        #     raise DependenciesError("Not all dependency-funcs are _DepFunc instances: %s"%bad_funcs, bad_funcs)
    except DependenciesError:
        raise
    except Exception as ex:
        raise DependenciesError("Bad explicit func_relations(%s) (item not a string, or deps not a tuple, etc) due to: %s"%(func_rels, ex), func_rels) from ex


def _build_func_dependencies_graph(func_rels, graph = None):
    if graph is None:
        graph = nx.DiGraph()

    func_rels = _consolidate_relations(func_rels)

    for (path, (deps, funcs)) in func_rels.items():
        if (deps):
            deps = _filter_common_prefixes(deps)
            graph.add_edges_from([(path, dep, {'funcs': funcs}) for dep in deps])

    cycles = list(nx.simple_cycles(graph))
    if cycles:
        log.warning('Cyclic dependencies! %s', cycles)

    return graph


def _consolidate_relations(relations):
    '''(item1, deps, func), (item1, ...) --> {item1, (set(deps), set(funcs))}'''

    rels = defaultdict()
    rels.default_factory = lambda: (set(), set()) # (deps, funcs)

    ## Join all item's  deps & funcs, and strip root-name.
    #
    for (item, deps, func) in relations:
        (pdes, pfuncs) = rels[item[_root_len:]]
        pdes.update([d[_root_len:] for d in deps])
        pfuncs.add(func)

    return rels


def _filter_common_prefixes(deps):
    '''deps: not-empty set

    example::
        deps = ['a', 'a.b', 'b.cc', 'a.d', 'b', 'ac', 'a.c']
        res = _filter_common_prefixes(deps)
        assert res == ['a.b', 'a.c', 'a.d', 'ac', 'b.cc']
    '''

    deps = sorted(deps)
    (it1, it2) = it.tee(deps)
    s2 = next(it2)
    ndeps = []
    try:
        while(True):
            s1=next(it1)
            s2=next(it2)
            if s1+'.' != s2[:len(s1)+1]:
                ndeps.append(s1)
    except StopIteration:
        ndeps.append(s2)

    return ndeps


def _research_calculation_routes(graph, sources, dests):
    '''Find nodes reaching 'dests' but not 'sources'.

        sources: a list of nodes (existent or not) to search for all paths originating from them
        dests:   a list of nodes to search for all paths leading to them them
        return: a 2-tuple with the graph and its nodes topologically-ordered
    '''

    ## Remove unrelated dests already present in sources.
    #
    calc_out_nodes = set(dests)
    calc_out_nodes -= set(sources)

    calc_inp_nodes = set(graph.nbunch_iter(sources))

    ## Deps graph: all INPUT's deps broken
    #    To be used for planing functions_to_run.
    #
    deps_graph = graph.copy()
    deps_graph.remove_edges_from(deps_graph.out_edges(calc_inp_nodes))

    ## Data_to_be_calced: all INPUTs erased
    #    To be used for topological-sorting.
    #
    data_graph = graph.copy()
    data_graph.remove_nodes_from(calc_inp_nodes)
    
    ## FIXME: IS it NEEDED?
    isolates = nx.isolates(data_graph)
    log.debug('Isolates: %s', isolates)
    data_graph.remove_nodes_from(isolates)
    try:
        calc_nodes = set(list(calc_out_nodes) + _all_predecessors(data_graph, calc_out_nodes))
    except (KeyError, NetworkXError) as ex:
        unknown = [node for node in calc_out_nodes if node not in graph]
        raise DependenciesError('Unknown OUT-args(%s)!' % unknown, (graph, unknown)) from ex
    else:
        return (calc_inp_nodes, calc_out_nodes, calc_nodes, deps_graph)

def _all_predecessors(graph, nodes):
    return [k for node in nodes for k in nx.bfs_predecessors(graph, node).keys()]


def _find_calculation_order(graph, calc_nodes):
    subgraph = graph.subgraph(calc_nodes)
    ordered_calc_nodes = list(reversed(nx.topological_sort(subgraph)))

    return ordered_calc_nodes


def _find_missing_input(calc_inp_nodes, graph):
    '''Search for *tentatively* missing data.'''
    calc_inp_nodes = set(calc_inp_nodes) # for efficiency below
    missing_input_nodes = []
    for node in nx.dfs_predecessors(graph):
        if ( node not in calc_inp_nodes and graph.out_degree(node) == 0):
            missing_input_nodes.append(node)
    return missing_input_nodes


def _extract_funcs_from_edges(graph, ordered_nodes):
    # f=list(fs[0]['funcs'])[0]
    funcs = [f for (_, _, d) in graph.edges_iter(ordered_nodes, True) if d
            for f in d['funcs']] # a list of sets


    ## Remove duplicates and No-funcs whilist preserving order.
    #
    od = OrderedDict.fromkeys(funcs)
    od.pop(None, None)
    funcs = list(od)

    return funcs




def _wrap_standalone_func(func):
    return _DepFunc(func=func, is_funcs_factory=False)
def _wrap_funcs_factory(funcs_factory):
    return _DepFunc(func=funcs_factory, is_funcs_factory=True)
def _wrap_child_func(funcs_factory, child_index):
    if not isinstance(funcs_factory, _DepFunc):
        funcs_factory = _wrap_funcs_factory(funcs_factory)
    return _DepFunc(func=funcs_factory, _child_index=child_index)
class _DepFunc:
    '''A wrapper for functions explored for relations, optionally allowing them to form a hierarchy of factories and produced functions.

    It can be in 3 types:
        * 0, standalone function: args given to function invocation are used immediatelt,
        * 10, functions-factory: args are stored and will be used by the child-funcs returned,
        * 20, child-func: created internally, and no args given when invoked, will use parent-factory's args.

    Use factory methods to create one of the first 2 types:
        * pdcalc._wrap_standalone_func()
        * pdcalc._wrap_funcs_factory()
    '''
    TYPES = ['standalone', 'funcs_fact', 'child']

    def __init__(self, func, is_funcs_factory=False, _child_index=None):
        self.func = func
        if is_funcs_factory:            ## Factory
            self._type = 1
            self.child_funcs = None

            assert _child_index == None, self
        elif _child_index is not None:  ## Child
            self._type = 2
            self.child_index = _child_index

            assert func.is_funcs_factory(), self
        else:                           ## Standalone
            self._type = 0

            assert _child_index == None, self

        if not callable(func):
            raise DependenciesError('Cannot create a _DepFunc for a non-callable(%s)!'%func, func)


    def get_type(self):
        return _DepFunc.TYPES[self._type]

    def is_standalone_func(self):
        return self._type == 0
    def is_funcs_factory(self):
        return self._type == 1
    def is_child_func(self):
        return self._type == 2

    def reset(self):
        if self.is_funcs_factory():
            self.child_funcs = None
    def is_reset(self):
        assert self.is_funcs_factory(), self

        return self.child_funcs is not None

    def mockup_func_args(self, root=None):
        assert not self.is_child_func(), self

        if not root:
            root = _make_mock(name=_root_name)

        sig = inspect.signature(self.func)
        mocks = []
        for (name, param) in sig.parameters.items():
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                log.warning('Any dependencies from **%s will be ignored for %s!', name, self)
                break
            mock = _make_mock()
            mocks.append(mock)
            root.attach_mock(mock, name)
        return (root, mocks)


    def __call__(self, *args, **kwargs):
        if self.is_standalone_func():           ## Standalone
            return self.func(*args, **kwargs)

        elif self.is_funcs_factory():           ## Factory
            cfuncs = self.func(*args, **kwargs)
            if not cfuncs:
                raise DependenciesError('%s returned %s as child-functions!'%(self.func, cfuncs), self.func)
            self.child_funcs = cfuncs
            return [_DepFunc(func=self, _child_index=i) for i in range(len(cfuncs))]

        else:                                   ## Child
            parent_fact = self.func
            assert parent_fact.is_funcs_factory(), self

            ## Use new args only if parent has previously been reset.
            #
            if (args or kwargs) and not parent_fact.is_reset():
                parent_fact(*args, **kwargs) ## Ignore returned depfuncs, we are the children!

            cfunc = parent_fact.child_funcs[self.child_index]
            return cfunc()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        try:
            if (self.is_child_func()):
                return '_DepFunc<child>(%s, %s)'%(self.func.func, self.child_index)
            return '_DepFunc<%s>(%s)'%(self.get_type(), self.func)
        except:
            return '_DepFunc<BAD_STR>(%s)'%self.func


##############################
## User Utilities
##############################
##

def default_arg_paths_extractor(arg_name, arg, paths):
    '''
    Dig recursively objects for indices to build the `sources`/`dests` sequences for :meth:`Dependencies.build_plan()`

    It extracts indices recursively from maps and pandas
    The inner-ones paths are not added, ie ``df.some.key``, but not ``df`` or ``df.some``.

    .. Note:: for pandas-series their index (their infos-axis) gets appended only the 1st time
        (not if invoked in recursion, from DataFrame columns).

    :param list paths: where to add the extracted paths into
    '''
    try:
        for key in arg.keys():
            path = '%s.%s'%(arg_name, key)
            value = arg[key]
            if (isinstance(value, pd.Series)):
                paths.append(path) # do not recurse into series.
            else:
                default_arg_paths_extractor(path, value, paths)
    except (AttributeError, KeyError):
        paths.append(arg_name)


def tell_paths_from_named_args(named_args, arg_paths_extractor_func=default_arg_paths_extractor, paths=None):
    '''
    Builds the `sources` or the `dests` sequences params of :meth:`Dependencies.build_plan()` from a map of function-arguments

    :param named_args: an ordered map ``{param_name --> param_value}`` similar to that returned by
            ``inspect.signature(func).bind(*args).arguments: BoundArguments``.
            Use the utility :func:`name_all_func_args()` to generate such a map for some function.
    :return: the `paths` updated with all ''dotted.vars'' found
    '''

    if paths is None:
        paths = []
    for (name, arg) in named_args.items():
        arg_paths_extractor_func(name, arg, paths)

    return paths


def name_all_func_args(func, *args, **kwargs):
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)

    return bound_args.arguments



##############################
## Classes
##############################
##

class Dependencies:
    '''
    Discovers and stores the rough functions-relationships needed to produce ExecutionPlanner

    The relation-tuples are "rough" in the sense that they may contain duplicates etc
    requiring cleanup by _consolidate_relations().

    **Usage:**
    

    Use the `harvest_{XXX}()` methods or the `add_func_rel()` method
    to gather dependencies from functions and function-facts
    and then use the `build_plan()` method to freeze them into a `plan`.
    '''

    @classmethod
    def from_funcs_map(cls, funcs_map, deps = None):
        '''
        Factory method for building `Dependencies` by harvesting multiple functions and func_factories, at once

        :param funcs_map: a mapping or a sequence of pairs ``(what --> bool_or_null)`` with values:
                
                True
                    when `what` is a funcs_factory, a function returning a sequence of functions processing the data
                False
                    when `what` is a standalone_function, or
                None
                    when `what` is an explicit relation 3-tuple (item, deps, func) to be fed
                    directly to :func:`Dependencies.add_func_rel()`, where:
                    
                    item 
                        a ''dotted.varname'' string,
                    deps
                        a string or a sequence of ''dotted.varname'' strings, and
                    func
                        a standalone func or a funcs_factory as a 2-tuple (func, index)
                        
                    Both items and deps are dot-separated sequence of varnames, such as 'foo.bar'
            
        :param deps: if none, a new `Dependencies` instance is created to harvest relations
        :return: a new (even inherited) or an updated `Dependencies` instance

        .. Important:: All functions must accept exactly the same args, or else the results
                will be undetermined and it will most-likely scream on execute_funcs_map.
        '''

        if deps is None:
            deps = cls()

        if isinstance(funcs_map, Mapping):
            pairs = funcs_map.items()
        else:
            pairs = funcs_map

        for (func, is_factory) in list(pairs):
            try:
                if is_factory is None:          ## relation-tuple
                    funcs_map.pop(func)
                    deps.add_func_rel(*func)
                elif is_factory:                ## funcs_factory
                    deps.harvest_funcs_factory(func)
                else:                           ## standalone-function
                    deps.harvest_func(func)
            except DependenciesError as ex:
                raise
            except Exception as ex:
                raise DependenciesError("Failed harvesting dependencies for %s due to: %s"%(func, ex), func) from ex

        return deps


    def __init__(self):
        self._relation_tuples = []

    def harvest_funcs_factory(self, funcs_factory):
        root = _make_mock(name=_root_name)
        harvest_funcs_factory(funcs_factory, root=root, func_rels=self._relation_tuples)
        log.debug('DEPS collected(%i): %s', len(self._relation_tuples), self._relation_tuples)

    def harvest_func(self, func):
        root = _make_mock(name=_root_name)
        harvest_func(func, root=root, func_rels=self._relation_tuples)
        log.debug('DEPS collected(%i): %s', len(self._relation_tuples), self._relation_tuples)

    def add_func_rel(self, item, deps=None, func=None):
        '''
        :param str item: a dotted.var
        :param deps: a string or a sequence of dotted.vars
        :param func: a standalone func, or a funcs_factory as a 2-tuple (func, index)
        '''

        ## Prepare args and do some basic argument-checking.
        #
        try:
            if func is not None:
                if isinstance(func, tuple):
                    (funcs_factory, child_index) = func
                    func = _wrap_child_func(funcs_factory, child_index)
                else:
                    func = _wrap_standalone_func(func)
        except (TypeError, ValueError) as ex:
            raise DependenciesError("Failed adding explicit func_relation(%s) due to bad functions\n  Function must either be a standalone-function or a 2-tuple(funcS_factory, child_index)."%item) from ex
        try:
            item = _root_name + '.' + item
            if deps is None:
                deps = []
            else:
                if isinstance(deps, str):
                    deps = [deps]
                deps = [_root_name + '.' + dep for dep in deps]
        except TypeError as ex:
            raise DependenciesError("Failed adding explicit func_relation(%s) due to bad item or dependencies strings!"%item) from ex

        new_func_rels = []
        _append_func_relation(item, deps, func, new_func_rels)

        _validate_func_relations(new_func_rels)

        self._relation_tuples.extend(new_func_rels)

    def _build_deps_graph(self):
        graph = _build_func_dependencies_graph(self._relation_tuples)
        return graph

    def _make_empty_plan(self):
        return pd.Series(dict(calc_inp_nodes=[], calc_out_nodes=[], calc_nodes=[],
            missing_data=[] if DEBUG else None, deps_graph=[]))

    def build_plan(self, sources, dests):
        '''
        Builds an execution-plan that when executed with :func:`execute_plan()` it will produce `dests` from `sources`.

        Turns any stored-relations into a the full dependencies-graph then trims it
        only to those ''dotted.varname'' nodes reaching from `sources` to `dests`.

        :param sequence dests:   a list of ''dotted.varname''s that will be the outcome of the calculations
        :param sources: a list of ''dotted.varname''s (existent or not) that are assumed to exist
                when the execution wil start
        :return: the new `plan` that can be fed to :func:`execute_plan()`
        :rtype: pd.Series
        '''

        log.debug('EXISTING data(%i): %s', len(sources), sources)
        log.debug('REQUESTED data(%i): %s', len(dests), dests)

        plan        = self._make_empty_plan()
        plan.dests  = dests

        graph       = self._build_deps_graph()
        log.debug('GRAPH constructed(%i): %s', graph.size(), graph.edges(data=True))

        (calc_inp_nodes, calc_out_nodes, unordered_calc_nodes, deps_graph) = \
                                _research_calculation_routes(graph, sources, dests)
        plan.deps_graph     = deps_graph
        plan.calc_out_nodes = calc_out_nodes
        plan.calc_inp_nodes = calc_inp_nodes

        calc_nodes          = _find_calculation_order(graph, unordered_calc_nodes)
        plan.calc_nodes     = calc_nodes
        plan.deps           = deps_graph.edges(calc_nodes, data=True)

        if DEBUG:
            missing_inp_nodes       = _find_missing_input(calc_inp_nodes, deps_graph)
            plan.missing_inp_nodes  = missing_inp_nodes
        else:
            plan.missing_inp_nodes  = None

        funcs               = _extract_funcs_from_edges(deps_graph, calc_nodes)
        plan.funcs          = funcs

        return plan




##############################
## Top-level methods
##############################
##

## Used if no default-dependencies specified in the `calculation` decorator, below.
calculation_deps = Dependencies()

def calculation(deps = None):
    '''A decorator that extracts dependencies from a function.

    :param deps: if None, any :subscript:`2` extracted deps are added in the `calculation_deps` module-variable.

    Example::

        @calculation
        def foo(a, r):
            a['bc'] = a.aa
            r['r1'] = a.bb

        @calculation
        def bar(a, r):
            r['r2'] = a.bc

        plan    = calculation_deps.build_plan(['a.ab', 'a.bc', 'r'], ['r.r2'])
        planner.execute_funcs_map(plan, 1, 2)
    '''

    if deps is None:
        deps = calculation_deps

    def decorator(func):
        deps.harvest_func(func)
        return func
    return decorator


def execute_plan(plan, *args, **kwargs):
    results = []
    for func in plan.funcs:
        try:
            results.append(func(*args, **kwargs))
        except Exception as ex:
            raise DependenciesError("Failed executing %s due to: %s"%(func, ex), func) from ex

    return results


def execute_funcs_factory(funcs_fact, dests, *args, **kwargs):
    '''A one-off way to run calculations from a funcions-factory (see :func:`execute_funcs_map()`)'''
    return execute_funcs_map({funcs_fact: True}, dests, *args, **kwargs)

def execute_funcs_map(funcs_map, dests, *args, **kwargs):
    '''
    A one-off way to run calculations from a funcs_map as defined in :meth:`Dependencies:from_funcs_map()`.

    :param funcs_map: a dictionary similar to the one used by :meth:`Dependencies.from_funcs_map()`
    :param seq dests: what is required as the final outcome of the execution,
            ie: for the func-functory::

                 def some_funcs(foo, bar):
                     def func_1():
                         foo['a1'] = ...
                     ...
                     def func_n(): ...
                         foo['a2'] = ...
                         bar['b1'] = ...

                     return [func_1, ..., func_n]

            the `dests` may be::

                ['foo.a1', 'foo.a2', 'bar.b1']

    :param args: the actual args to use for invoking the functions in the map,
            and the names of these args would come rom the first function
            to be invoked (which ever that might be!).


    .. Note:: Do not use it to run the calculations repeatedly.  Preffer to cache the function-relations
            into an intermediate :class:`Dependencies` instance, instead.
    '''

    ## Find the first func in the map and
    #    use it for naming the args.
    #
    a_func = None
    for func in funcs_map.keys():
        if callable(func):
            a_func = func
            break
    else:
        raise DependenciesError('No function found in funcs_map(%s)!'%funcs_map, funcs_map)
    named_args  = name_all_func_args(a_func, *args, **kwargs)
    sources     = tell_paths_from_named_args(named_args)

    deps        = Dependencies.from_funcs_map(funcs_map)
    plan        = deps.build_plan(sources, dests)

    return execute_plan(plan, *args, **kwargs)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from ._version import (__version__, __updated__)


__copyright__     = "Copyright (C) 2014 European Commission (JRC)"
__license__       = "EUPL 1.1+"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
from unittest.case import skipIf
'''Check xlwings excel functionality.
'''

import unittest
import logging
import os, sys
import pandas as pd
import numpy as np
from numpy import testing as npt
from pandas.core.generic import NDFrame


DEFAULT_LOG_LEVEL   = logging.INFO
def _init_logging(loglevel):
    logging.basicConfig(level=loglevel)
    logging.getLogger().setLevel(level=loglevel)

    log = logging.getLogger(__name__)
    log.trace = lambda *args, **kws: log.log(0, *args, **kws)
    
    return log
log = _init_logging(DEFAULT_LOG_LEVEL)


def from_my_path(*parts):
    return os.path.join(os.path.dirname(__file__), *parts)

def _make_sample_workbook(sheetname, addr, value):
    import xlwings as xw
    wb = xw.Workbook()
    xw.Sheet(1).name = sheetname
    xw.Range(addr).value = value

    return wb

def close_workbook(wb):
    try:
        wb.close()
    except Exception:
        log.warning('Minor failure while closing Workbook!', exc_info=True)


@skipIf(not ('darwin' in sys.platform or 'win32' in sys.platform), "Cannot test xlwings in Linux")
class TestExcel(unittest.TestCase):


    def test_xlwings_smoketest(self):
        import xlwings as xw
        sheetname   ='shitt'
        addr        ='f6'
        table       = pd.DataFrame([[1,2],[True, 'off']], columns=list('ab'))
        wb = xw.Workbook()
        try:
            xw.Sheet(1).name = sheetname
            xw.Range(addr).value = table
        finally:
            close_workbook(wb)


    def test_excel_refs(self):
        from ..excel.FuefitExcelRunner import resolve_excel_ref
        sheetname   ='Input'
        addr        ='d2'
        table       = pd.DataFrame({'a':[1,2,3], 'b':['s','t','u'], 'c':[True,False,True]})
        table.index.name = 'id'
        
        cases = [
            ("@d2",                         'id'),
            ("@e2",                         'a'),
            ("@e3",                         1),
            ("@e3:g3",                      table.iloc[0, :].values.reshape((3,1))),
            ("@1!e2.horizontal",            table.columns.values.reshape((3,1))),
            ("@1!d3.vertical",              table.index.values.reshape((3,1))),
            ("@e4.horizontal",              table.iloc[1, :].values.reshape((3,1))),
            ("@1!e3:g3.vertical",           table.values),
            ("@{sheet}!E2:F2.table(strict=True, header=True)".format(sheet=sheetname),
                                            table),
        ]
        
        errors = []
        wb = _make_sample_workbook(sheetname, addr, table)
        try:
            for i, (inp, exp) in enumerate(cases):
                out = None
                try:
                    out = resolve_excel_ref(inp)
                    log.debug('%i: INP(%s), OUT(%s), EXP(%s)', i, inp, out, exp)
                    if not exp is None:
                        if isinstance(out, NDFrame):
                            out = out.values
                        if isinstance(exp, NDFrame):
                            exp = exp.values
                        if isinstance(out, np.ndarray) or isinstance(exp, np.ndarray):
                            npt.assert_array_equal(out, exp, '%i: INP(%s), OUT(%s), EXP(%s)' % (i, inp, out, exp))
                        else:
                            self.assertEqual(out, exp, '%i: INP(%s), OUT(%s), EXP(%s)' % (i, inp, out, exp))
                except Exception as ex:
                    log.exception('%i: INP(%s), OUT(%s), EXP(%s), FAIL: %s' % (i, inp, out, exp, ex))
                    errors.append(ex)
        finally:
            close_workbook(wb)

        if errors:
            raise Exception('There are %i out of %i errors!'% (len(errors), len(cases)))


    def test_excel_runner_call_from_python(self):
        import xlwings as xw
        from ..excel import FuefitExcelRunner 
        wb = xw.Workbook(from_my_path('..', 'excel', 'FuefitExcelRunner'))
        try:
            FuefitExcelRunner.main()
        finally:
            close_workbook(wb)


if __name__ == "__main__":
    unittest.main()



#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
## Authoritative project's version-info
__version__ = "0.0.5"
__updated__ = "2014-11-12 16:22:55"



import contextlib
import os
import sys

## Adapted from the response to issue(http://bugs.python.org/issue15805):
#    https://gist.githubusercontent.com/msabramo/6043474/raw/b02894848fd683b556f27c0ac2dbcf316f538168/redirect.py
#

@contextlib.contextmanager
def redirected(**kwargs):
    """
    A context manager to temporarily redirect stdout or stderr

    Examples:

    # Redirect stdout to /dev/null
    >>> with redirected(stdout=None):
    ...     os.system("echo foo; ls dfkdjfdkfd")
    ...
    ls: dfkdjfdkfd: No such file or directory

    # Redirect stderr to /dev/null
    >>> with redirected(stderr=None):
    ...     os.system("echo foo; ls djdffdd")
    ...
    foo
    256

    # Redirect stdout and stderr to /dev/null
    >>> with redirected(stdout=None, stderr=None):
    ...     os.system("echo foo; ls djdffdd")
    ...

    # Redirect stdout and stderr to filenames
    >>> with redirected(stdout='stuff_stdout.txt', stderr='stuff_stderr.txt'):
    ...     ret = os.system("echo foo; ls dfjkdfd")
    ...
    >>> open('stuff_stdout.txt').read()
    'foo\n'
    >>> open('stuff_stderr.txt').read()
    'ls: dfjkdfd: No such file or directory\n'

    # Redirect stdout to a file and stderr to a filename
    >>> with open('going_to_stdout.txt', 'w') as out:
    ...     with redirected(stdout=out, stderr='stuff_stderr.txt'):
    ...         ret = os.system("echo stuff going to stdout; ls 123456")
    ...
    >>> open('going_to_stdout.txt', 'r').read()
    'stuff going to stdout\n'
    >>> open('stuff_stderr.txt', 'r').read()
    'ls: 123456: No such file or directory\n'

    # Redirect stdout to a file-like object
    # This does not work with subprocesses.
    >>> from StringIO import StringIO
    >>> out = StringIO()
    >>> with redirected(stdout=out):
    ...     print('printing some stuff')
    ...
    >>> out.getvalue()
    'printing some stuff\n'

    """

    dest_files = {}
    old_filenos = {}
    old_channels = {}

    try:
        for channel_name, destination in kwargs.items():
            stdchannel = getattr(sys, channel_name)
            try:
                old_filenos[channel_name] = os.dup(stdchannel.fileno())
                dupped = True
            except:
                # # Failback if not a real file (i.e. Eclipse's console).
                #
                old_channels[channel_name] = stdchannel
                dupped = False

            if destination is None:
                dest_file = open(os.devnull, 'w')
            elif hasattr(destination, 'startswith'):  # A string => treat as filename
                dest_file = open(destination, 'w')
            else:
                try:
                    _ = destination.fileno()  # A file-like object
                    dest_file = destination
                except Exception:
                    dest_file = None

            if dest_file and dupped:
                os.dup2(dest_file.fileno(), stdchannel.fileno())
                dest_files[channel_name] = dest_file
            else:
                setattr(sys, channel_name, destination)

        yield
    finally:
        for channel_name, old_channel in old_channels.items():
            setattr(sys, channel_name, old_channel)

        for channel_name, old_fileno in old_filenos.items():
            setattr(sys, '%s' % channel_name, getattr(sys, '__%s__' % channel_name))

            if channel_name in dest_files:
                stdchannel = getattr(sys, channel_name)
                dest_file = dest_files[channel_name]

                if old_fileno is not None:
                    os.dup2(old_fileno, stdchannel.fileno())
                if dest_file is not None:
                    dest_file.close()




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
''''
fuefit: Fit fuel-consumption engine-maps on a physical formula with 6 parameters.


Install:
========

To install it, assuming you have download the sources,
do the usual::

    python setup.py install

Or get it directly from the PIP repository::

    pip install fuefit


Tested with Python 3.4.
'''

from distutils.version import StrictVersion
import os, sys, io
import re

from setuptools import setup


## Fail early in ancient python-versions
#
py_verinfo = sys.version_info
py_sver = StrictVersion("%s.%s.%s" % py_verinfo[:3])
if py_verinfo[0] != 3 or py_sver < StrictVersion("3.3"):
    exit("Sorry, only Python >= 3.3 is supported!")
if sys.argv[-1] == 'setup.py':
    exit("To install, run `python setup.py install`")
    

proj_name = 'fuefit'
mydir = os.path.dirname(__file__)

## Version-trick to have version-info in a single place,
## taken from: http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
##
def read_project_version():
    fglobals = {}
    with io.open(os.path.join(mydir, proj_name, '_version.py')) as fd:
        exec(fd.read(), fglobals)  # To read __version__
    return fglobals['__version__']

def read_text_lines(fname):
    with io.open(os.path.join(mydir, fname)) as fd:
        return fd.readlines()

def yield_sphinx_only_markup(lines):
    """
    :param file_inp:     a `filename` or ``sys.stdin``?
    :param file_out:     a `filename` or ``sys.stdout`?`

    """
    substs = [
        ## Selected Sphinx-only Roles.
        #
        (r':abbr:`([^`]+)`',        r'\1'),
        (r':ref:`([^`]+)`',         r'`\1`_'),
        (r':term:`([^`]+)`',        r'**\1**'),
        (r':dfn:`([^`]+)`',         r'**\1**'),
        (r':(samp|guilabel|menuselection):`([^`]+)`',        r'``\2``'),


        ## Sphinx-only roles:
        #        :foo:`bar`   --> foo(``bar``)
        #        :a:foo:`bar` XXX afoo(``bar``)
        #
        #(r'(:(\w+))?:(\w+):`([^`]*)`', r'\2\3(``\4``)'),
        (r':(\w+):`([^`]*)`', r'\1(``\2``)'),


        ## Sphinx-only Directives.
        #
        (r'\.\. doctest',           r'code-block'),
        (r'\.\. plot::',            r'.. '),
        (r'\.\. seealso',           r'info'),
        (r'\.\. glossary',          r'rubric'),
        (r'\.\. figure::',          r'.. '),
        (r'\.\. image::',          r'.. '),


        ## Other
        #
        (r'\|version\|',              r'x.x.x'),
    ]

    regex_subs = [ (re.compile(regex, re.IGNORECASE), sub) for (regex, sub) in substs ]

    def clean_line(line):
        try:
            for (regex, sub) in regex_subs:
                line = regex.sub(sub, line)
        except Exception as ex:
            print("ERROR: %s, (line(%s)"%(regex, sub))
            raise ex

        return line

    for line in lines:
        yield clean_line(line)



proj_ver = read_project_version()


readme_lines = read_text_lines('README.rst')
description = readme_lines[1]
long_desc = ''.join(yield_sphinx_only_markup(readme_lines))
## Trick from: http://peterdowns.com/posts/first-time-with-pypi.html
download_url = 'https://github.com/ankostis/%s/tarball/v%s' %(proj_name, proj_ver)

setup(
    name=proj_name,
    version=proj_ver,
    description=description,
    long_description=long_desc,
    author="Kostis Anagnostopoulos @ European Commission (JRC)",
    author_email="ankostis@gmail.com",
    url="https://github.com/ankostis/fuefit",
    download_url=download_url,
    license = "European Union Public Licence 1.1 or later (EUPL 1.1+)",
    keywords = [
         "automotive", "vehicle", "vehicles", "car", "cars", "fuel", "consumption",
        "engine", "engine-map", "fitting",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Development Status :: 4 - Beta",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=['fuefit', 'fuefit.test', 'fuefit.excel'],
    package_data={
        'fuefit.test': ['*.bat', '*.sh'],
        'fuefit.excel': ['*.ico'],
    },
#    extras_require = {
#        'Docs':  ['sphinx >= 1.2'],
#    },
    include_package_data = True,
    install_requires=[
        'enum34',
        'pandas',
        'xlrd',
        'scipy',
        'lmfit',
        'jsonschema',
        'networkx',
        'pint',
        'xlwings >= 0.2.3',
    ],
    setup_requires = [
        'setuptools',# >= 3.4.4',
        'setuptools-git >= 0.3', ## Gather package-data from all files in git.
        'nose>=1.0',
        'sphinx', # >=1.3',
        'sphinx_rtd_theme',
        'matplotlib',
        'wheel',
    ],
    tests_require = [
        'nose',
    ],
    test_suite='nose.collector',
    entry_points={
        'console_scripts': [
            'fuefit          = fuefit.__main__:main',
        ],
    },
    zip_safe=False,
    options={
        'build_sphinx' :{
            'build_dir': 'Docs/_build',
        },
        'bdist_wheel' :{
            'universal': True,
        },
    },
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
Check pdcalc's function-dependencies exploration, reporting and classes .
'''
from collections import OrderedDict
import logging
import unittest

from networkx.classes.digraph import DiGraph

import pandas as pd

from ..pdcalc import (
    DependenciesError, execute_funcs_map, execute_plan,
    Dependencies, _research_calculation_routes,
    tell_paths_from_named_args
)


def lstr(lst):
    return '\n'.join([str(e) for e in lst])

def DF(d):
    return pd.DataFrame(d)
def SR(d):
    return pd.Series(d)


def make_test_graph():
    '''     (4)  6
             ^   ^
             | X |
            (3)  5
             ^   ^
              \ /
              [2]
               ^
               |
              (1)
    '''
    web = DiGraph()
    web.add_edges_from([(1,2), (2,3), (3,4), (2,5), (3,6), (5,6), (5,4)])
    return web


def funcs_fact1(params, engine, dfin, dfout):
    from math import pi

    def f1(): engine['fuel_lhv'] = params['fuel'][engine['fuel']]['lhv']
    def f2(): dfin['n']     = dfin.n_norm * (engine.n_rated - engine.n_idle) + engine.n_idle
    def f3(): dfin['p']       = dfin.p_norm * engine.p_max
    def f4(): dfin['fc']      = dfin.fc_norm * engine.p_max
    def f5(): dfin['rps']     = dfin.n / 60
    def f6(): dfin['torque']  = (dfin.p * 1000) / (dfin.rps * 2 * pi)
    def f7(): dfin['pme']     = (dfin.torque * 10e-5 * 4 * pi) / (engine.capacity * 10e-6)
    def f8(): dfin['pmf']     = ((4 * pi * engine.fuel_lhv) / (engine.capacity * 10e-3)) * (dfin.fc / (3600 * dfin.rps * 2 * pi)) * 10e-5
    def f9(): dfin['cm']      = dfin.rps * 2 * engine.stroke / 1000
    return (f1, f2, f3, f4, f5, f6, f7, f8, f9)

def funcs_fact2(params, engine, dfin, dfout):
    ## Out of returned funcs!!
    def f10(): return dfin.cm + dfin.pmf + dfin.pme

    def f11(): engine['eng_map_params'] = f10()
    def f12():
        dfout['n']    = engine['eng_map_params']
        dfout['p']      = engine['eng_map_params'] * 2
        dfout['fc']     = engine['eng_map_params'] * 4
    def f13(): dfout['fc_norm']         = dfout.fc / dfout.p

    return (f11, f12, f13)

def funcs_fact3(params, engine, dfin, dfout):
    def f12():
        dfout['n']    = engine['eng_map_params']
        dfout['p']      = engine['eng_map_params'] * 2
        dfout['fc']     = engine['eng_map_params'] * 4
    def f13(): dfout['fc_norm']         = dfout.fc / dfout.p

    return (f12, f13)


def func11(params, engine, dfin, dfout):
    engine['eng_map_params'] = dfin.cm + dfin.pmf + dfin.pme



def funcs_fact(params, engine, dfin, dfout):
    return funcs_fact1(params, engine, dfin, dfout) + funcs_fact2(params, engine, dfin, dfout)

def get_params():
    return {
        'fuel': {
            'diesel':{'lhv':42700},
            'petrol':{'lhv':43000}
        },
    }

def get_engine():
    return {
        'fuel': 'diesel',
#         'measured_eng_points': '',
        'n_idle': 100,
        'n_rated' : 12,
        'p_max'     : 2,
        'capacity'  : 1300,
        'stroke'    : 12,
    }

def build_base_deps():
    deps = Dependencies()
    deps.harvest_funcs_factory(funcs_fact)

    return deps

def make_plan_and_execute(deps, dests, named_args=None, sources=None):
    if not named_args is None:
        sources = tell_paths_from_named_args(named_args)
    elif sources is None:
        raise Exception("One of `named_args`, `sources` should not be None!")

    plan = deps.build_plan(sources, dests)
    #log.info('Execution PLAN: %s', plan)
    return execute_plan(plan, *named_args.values())


def bad_func(params, engine, dfin, dfout):
    raise AssertionError('Bad func invoked!')
def bad_funcs_fact(params, engine, dfin, dfout):
    def child():
        raise AssertionError("Bad funcs_fact's child invoked!")
    return (child, )


class Test(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.DEBUG)
        l=logging.getLogger()
        l.setLevel(logging.DEBUG)
        l.handlers = [logging.StreamHandler()]



    def testSmoke_FuncExplorer_countNodes(self):
        deps = build_base_deps()
        graph = deps._build_deps_graph()
#         print("RELS:\n", lstr(deps.rels))
#         print('ORDERED:\n', lstr(plan.graph.ordered(True)))
        self.assertEqual(len(graph), 25) #29 when adding.segments

        return graph

    def test_find_connecting_nodes_smoke(self):
        deps = build_base_deps()
        graph = deps._build_deps_graph()

        inp = ('dfin.fc', 'dfin.fc_norm', 'dfin.XX')
        out = ('dfout.fc', 'dfout.n')
        (_, _, cn_nodes, _) = _research_calculation_routes(graph, inp, out)
        self.assertTrue('dfin.fc_norm' not in cn_nodes)

        #print(cn_nodes)

    def test_find_connecting_nodes_fail(self):
        deps = build_base_deps()
        graph = deps._build_deps_graph()

        inp = ('dfin.fc', 'dfin.fc_norm')
        out = ('dfout.fc', 'dfout.BAD')
        with self.assertRaisesRegex(DependenciesError, 'dfout\.BAD'):
            _research_calculation_routes(graph, inp, out)


    def test_find_connecting_nodes_good(self):
        web = make_test_graph()
        web.remove_edges_from([(5,6), (5,4)])

        inp = (1, 3, 4)
        out = (2,)
        (_, _, cn_nodes, _) = _research_calculation_routes(web, inp, out)
        all_out = {1,4,3}
        all_in = {2,5}
        self.assertTrue(cn_nodes - all_out == cn_nodes, cn_nodes)
        self.assertTrue(all_in & cn_nodes == all_in, cn_nodes)

    def test_find_connecting_nodes_good_sharedDep(self):
        web = make_test_graph()
        web.remove_edges_from([(5,4)])

        inp = (1, 3, 4)
        out = (2,)
        (_, _, cn_nodes, _) = _research_calculation_routes(web, inp, out)
        all_out = {1,4,3}
        all_in = {2,5,6}
        self.assertTrue(cn_nodes - all_out == cn_nodes, cn_nodes)
        self.assertTrue(all_in & cn_nodes == all_in, cn_nodes)

    def test_find_connecting_nodes_good_sourceDep(self):
        '''     (4)  6
                 ^   ^
                 | X |
                (3)  5
                 ^   ^
                  \ /
                  [2]
                   ^
                   |
                  (1)
        '''
        web = DiGraph()
        web.add_edges_from([(1,2), (2,3), (3,4), (2,5), (3,6), (5,6), (5,4)])

        inp = (1, 3, 4)
        out = (2,)
        (_, _, cn_nodes, _) = _research_calculation_routes(web, inp, out)
        all_out = {1,4,3}
        all_in = {2,5,6}
        self.assertTrue(cn_nodes - all_out == cn_nodes, cn_nodes)
        self.assertTrue(all_in & cn_nodes == all_in, cn_nodes)


    def testSmoke_ExecutionPlan_fail(self):
        deps = build_base_deps()

        args = {}
        inp = ('dfin.fc', 'dfin.fc_norm')
        out = ('dfout.fc', 'dfout.BAD')
        with self.assertRaisesRegex(DependenciesError, 'dfout\.BAD'):
            make_plan_and_execute(deps, out, args, inp)

    def test_tell_paths_from_named_args_dicts(self):
        d = {'arg1':{'a':1, 'b':2}, 'arg2':{11:11, 12:{13:13}}}

        paths = tell_paths_from_named_args(d)
        self.assertTrue('arg1.a' in paths, paths)
        self.assertFalse('arg1.a.1' in paths, paths)
        self.assertTrue('arg1.b' in paths, paths)
        self.assertFalse('arg1.b.2' in paths, paths)
        self.assertTrue('arg2.11' in paths, paths)
        self.assertFalse('arg2.11.2' in paths, paths)
        self.assertTrue('arg2.12.13' in paths, paths)
        self.assertFalse('arg2.12.13.13' in paths, paths)

    def test_tell_paths_from_named_args_DF(self):
        d = pd.DataFrame({'a':[1,2], 'b':[3,4]})

        paths = tell_paths_from_named_args({'arg1':d})
        self.assertTrue('arg1.a' in paths, paths)
        self.assertTrue('arg1.b' in paths, paths)
        self.assertEqual(len(paths), 2, paths)

    def test_tell_paths_from_named_args_Series(self):
        d = pd.Series({'a':[1,2], 'b':3})

        paths = tell_paths_from_named_args({'arg1':d})
        self.assertTrue('arg1.a' in paths, paths)
        self.assertTrue('arg1.b' in paths, paths)
        self.assertEqual(len(paths), 2, paths)


    def testSmoke_ExecutionPlan_good(self):
        deps = build_base_deps()

        ## TODO, Check dotted.var.names.
        engine = SR(get_engine())
        dfin = DF({'fc':[1, 2], 'fc_norm':[22, 44], 'n':[10,20], 'pme':[100,200], 'some_foo':[1,2]})
        dfout = DF({})
        args = OrderedDict([
            ('params', SR(get_params())),
            ('engine', engine),
            ('dfin',  dfin),
            ('dfout', dfout),
        ])
        out = ('dfout.n', 'dfout.fc_norm')

        engine_c = engine.copy()
        dfin_c = dfin.copy()
        dfout_c = dfout.copy()

        make_plan_and_execute(deps, out, named_args=args)

        ## Check args modified!
        self.assertFalse(engine.equals(engine_c), engine)
        self.assertFalse(dfin.equals(dfin_c), dfin)
        self.assertFalse(dfout.equals(dfout_c), dfout)

    def testSmoke_ExecutionPlan_goodExtraRels(self):
        deps = build_base_deps()
        deps.add_func_rel('engine.fuel_lhv', ('params.fuel.diesel.lhv', 'params.fuel.petrol.lhv'))

        engine = SR(get_engine())
        dfin = DF({'fc':[1, 2], 'fc_norm':[22, 44], 'n':[10,20], 'pme':[100,200], 'some_foo':[1,2]})
        dfout = DF({})
        args = OrderedDict([
            ('params', SR(get_params())),
            ('engine', engine),
            ('dfin',  dfin),
            ('dfout', dfout),
        ])
        out = ('dfout.n', 'dfout.fc_norm')

        engine_c = engine.copy()
        dfin_c = dfin.copy()
        dfout_c = dfout.copy()

        make_plan_and_execute(deps, out, named_args=args)

        ## Check args modified!
        self.assertFalse(engine.equals(engine_c), engine)
        self.assertFalse(dfin.equals(dfin_c), dfin)
        self.assertFalse(dfout.equals(dfout_c), dfout)


    def testSmoke_ExecutionPlan_multiFatcs_good(self):
        deps = Dependencies()
        deps.harvest_funcs_factory(funcs_fact1)
        deps.harvest_funcs_factory(funcs_fact2)
        deps.add_func_rel('engine.fuel_lhv', ('params.fuel.diesel.lhv', 'params.fuel.petrol.lhv'))

        engine = SR(get_engine())
        dfin = DF({'fc':[1, 2], 'fc_norm':[22, 44], 'n':[10,20], 'pme':[100,200], 'some_foo':[1,2]})
        dfout = DF({})
        args = OrderedDict([
            ('params', SR(get_params())),
            ('engine', engine),
            ('dfin',  dfin),
            ('dfout', dfout),
        ])
        out = ('dfout.n', 'dfout.fc_norm')

        engine_c = engine.copy()
        dfin_c = dfin.copy()
        dfout_c = dfout.copy()

        make_plan_and_execute(deps, out, args)

        ## Check args modified!
        self.assertFalse(engine.equals(engine_c), engine)
        self.assertFalse(dfin.equals(dfin_c), dfin)
        self.assertFalse(dfout.equals(dfout_c), dfout)


    def testSmoke_funcs_map_good(self):
        params = SR(get_params())
        engine = SR(get_engine())
        dfin = DF({'fc':[1, 2], 'fc_norm':[22, 44], 'n':[10,20], 'pme':[100,200], 'some_foo':[1,2]})
        dfout = DF({})
        out = ('dfout.n', 'dfout.fc_norm')

        engine_c = engine.copy()
        dfin_c = dfin.copy()
        dfout_c = dfout.copy()


        funcs_map = {
            funcs_fact1: True,
            funcs_fact3: True,
            func11: False,
            ('engine.fuel_lhv', ('params.fuel.diesel.lhv', 'params.fuel.petrol.lhv'), None): None,
            ('a.standalone', 'some.dep', bad_func): None,
            ('a.funcs_fact', ('one.dep', 'another.dep'), (bad_funcs_fact, 0)): None,
        }

        execute_funcs_map(funcs_map, out, params, engine, dfout=dfout, dfin=dfin)

        ## Check args modified!
        self.assertFalse(engine.equals(engine_c), engine)
        self.assertFalse(dfin.equals(dfin_c), dfin)
        self.assertFalse(dfout.equals(dfout_c), dfout)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
